# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import functools
import itertools
from typing import NamedTuple, Sequence, Optional, Type

import pytest

import tests

import amazon.ion.symbols as symbols


def test_system_symbols():
    sys_table = symbols.SYSTEM_SYMBOL_TABLE

    assert 9 == len(sys_table)
    assert u'$ion' == sys_table.name
    assert 1 == sys_table.version

    with pytest.raises(TypeError):
        sys_table.intern(u'moo')

    with pytest.raises(ValueError):
        symbols.SymbolTable(
            table_type=symbols.SYSTEM_TABLE_TYPE,
            symbols=(),
            name=u'my_ion',
            version=1
        )

    with pytest.raises(ValueError):
        symbols.SymbolTable(
            table_type=symbols.SYSTEM_TABLE_TYPE,
            symbols=(),
            name=u'$ion',
            version=2,
            imports=sys_table
        )


class _P(NamedTuple):
    name: Optional[str]
    version: Optional[int]
    symbols: Sequence[str]
    exc: Type[Exception] = ValueError

    def __str__(self):
        return '{p.name!r}, {p.version}, {p.symbols!r}'.format(p = self)


@tests.parametrize(
    _P(name=None, version=None, symbols=[]),
    _P(name=None, version=1, symbols=[]),
    _P(name=u'my_shared', version=None, symbols=[]),
    _P(name=u'my_shared', version=0, symbols=[]),
    _P(name=u'my_shared', version=-1, symbols=[]),
    _P(name=b'my_shared', version=1, symbols=[], exc=TypeError),
    _P(name=u'my_shared', version=1, symbols=[b'a'], exc=TypeError),
)
def test_shared_symbols_malformed(p):
    with pytest.raises(p.exc):
        symbols.shared_symbol_table(p.name, p.version, p.symbols)


FOO_TEXTS = (u'a', u'b', u'c')
FOO_TABLE = symbols.shared_symbol_table(
    name=u'foo',
    version=1,
    symbols=FOO_TEXTS
)

BAR_NEW_TEXTS = (u'c', u'd', u'e')
BAR_TEXTS = FOO_TEXTS + BAR_NEW_TEXTS
BAR_TABLE = symbols.shared_symbol_table(
    name=u'bar',
    version=2,
    symbols=BAR_NEW_TEXTS,
    imports=(FOO_TABLE,)
)

PLACEHOLDER = symbols.placeholder_symbol_table(u'placeholder', 1, 10)
PLACEHOLDER_TEXTS = tuple(itertools.repeat(None, 10))

SUB_SOURCE_LESS_SYMBOLS = symbols.substitute_symbol_table(BAR_TABLE, 3, 10)
SUB_SOURCE_LESS_TEXTS = tuple(itertools.chain(BAR_TEXTS, itertools.repeat(None, 4)))

SUB_SOURCE_MORE_SYMBOLS = symbols.substitute_symbol_table(BAR_TABLE, 1, 4)
SUB_SOURCE_MORE_TEXTS = BAR_TEXTS[:4]

SUB_SOURCE_EQUAL_SYMBOLS = symbols.substitute_symbol_table(BAR_TABLE, 2, 6)
SUB_SOURCE_EQUAL_TEXTS = BAR_TEXTS


class _P(NamedTuple):
    desc: str
    table: symbols.SymbolTable
    name: str
    version: int
    is_substitute: bool
    symbol_texts: Sequence[Optional[str]]

    def __str__(self):
        return  self.desc


@tests.parametrize(
    _P(
        desc='FOO_TABLE',
        table=FOO_TABLE,
        name=u'foo',
        version=1,
        is_substitute=False,
        symbol_texts=FOO_TEXTS
    ),
    _P(
        desc='BAR_TABLE',
        table=BAR_TABLE,
        name=u'bar',
        version=2,
        is_substitute=False,
        symbol_texts=BAR_TEXTS
    ),
    _P(
        desc='PLACEHOLDER',
        table=PLACEHOLDER,
        name=u'placeholder',
        version=1,
        is_substitute=True,
        symbol_texts=PLACEHOLDER_TEXTS
    ),
    _P(
        desc='SUB_SOURCE_LESS',
        table=SUB_SOURCE_LESS_SYMBOLS,
        name=u'bar',
        version=3,
        is_substitute=True,
        symbol_texts=SUB_SOURCE_LESS_TEXTS
    ),
    _P(
        desc='SUB_SOURCE_MORE',
        table=SUB_SOURCE_MORE_SYMBOLS,
        name=u'bar',
        version=1,
        is_substitute=True,
        symbol_texts=SUB_SOURCE_MORE_TEXTS
    ),
    _P(
        desc='SUB_SOURCE_EQUALS',
        table=SUB_SOURCE_EQUAL_SYMBOLS,
        name=u'bar',
        version=2,
        is_substitute=True,
        symbol_texts=SUB_SOURCE_EQUAL_TEXTS),
)
def test_shared_symbols(p):
    assert p.name == p.table.name
    assert p.version == p.table.version
    assert len(p.symbol_texts) == len(p.table)
    assert p.is_substitute == p.table.is_substitute

    text_iter = iter(p.symbol_texts)
    curr_sid = 1
    for token in p.table:
        text = next(text_iter)
        assert text == token.text
        assert curr_sid == token.sid
        assert symbols.ImportLocation(p.name, curr_sid) == token.location
        curr_sid += 1

    assert None == p.table.get(u'z')
    with pytest.raises(KeyError):
        p.table[u'z']
    assert None == p.table.get(1024)
    with pytest.raises(KeyError):
        p.table[1024]

    seen = set()
    curr_id = 1
    for text in p.symbol_texts:
        token = p.table[curr_id]
        assert text == token.text
        if text is not None:
            if text not in seen:
                seen.add(text)
                assert token == p.table[text]
            else:
                mapped_token = p.table[text]
                assert token != mapped_token
                assert token.sid > mapped_token.sid
        curr_id += 1

    assert symbols.SYMBOL_ZERO_TOKEN == p.table[0]


def test_shared_symbols_intern():
    with pytest.raises(TypeError):
        FOO_TABLE.intern(u'hello')


class _P(NamedTuple):
    name: Optional[str]
    version: Optional[int]

    def __str__(self):
        return '{p.name!r} {p.version!r}'.format(p=self)


@tests.parametrize(
    _P(u'local_name', None),
    _P(None, 1),
    _P(u'local_name', 1)
)
def test_local_symbols_malformed(p):
    with pytest.raises(ValueError):
        symbols.SymbolTable(
            table_type=symbols.LOCAL_TABLE_TYPE,
            symbols=[u'something'],
            name=p.name,
            version=p.version
        )


COLLIDE_TABLE = symbols.shared_symbol_table(
    name=u'collide',
    version=3,
    symbols=tuple(x.text for x in symbols._SYSTEM_SYMBOL_TOKENS)
)

_T = symbols.SymbolToken
_L = functools.partial(symbols.ImportLocation, FOO_TABLE.name)
_SID_START = len(symbols.SYSTEM_SYMBOL_TABLE) + 1
_COLLIDE_SID_START = len(symbols.SYSTEM_SYMBOL_TABLE) + len(COLLIDE_TABLE) + 1


class _P(NamedTuple):
    desc: str
    symbol_texts: Sequence[str]
    tokens: Sequence[symbols.SymbolToken]
    imports: Sequence = ()
    symbol_count_override: Optional[int] = None

    def __str__(self):
        return self.desc


@tests.parametrize(
    _P(
        desc='DISTINCT',
        symbol_texts=[u'a', u'b', u'c'],
        tokens=[
            _T(u'a', _SID_START + 0),
            _T(u'b', _SID_START + 1),
            _T(u'c', _SID_START + 2),
        ],
    ),
    _P(
        desc='DUPLICATE',
        symbol_texts=[u'a', u'a', u'c'],
        tokens=[
            _T(u'a', _SID_START + 0),
            _T(u'c', _SID_START + 1),
        ],
    ),
    _P(
        desc='DUPLICATE SYSTEM',
        symbol_texts=[u'name', u'version', u'c'],
        tokens=[
            _T(u'c', _SID_START + 0),
        ],
    ),
    _P(
        desc='IMPORT DISTINCT',
        symbol_texts=[u'x', u'y', u'z'],
        imports=[
            FOO_TABLE,
        ],
        tokens=[
            _T(u'a', _SID_START + 0, _L(1)),
            _T(u'b', _SID_START + 1, _L(2)),
            _T(u'c', _SID_START + 2, _L(3)),
            _T(u'x', _SID_START + 3),
            _T(u'y', _SID_START + 4),
            _T(u'z', _SID_START + 5),
        ],
    ),
    _P(
        desc='IMPORT DUP',
        symbol_texts=[u'c', u'y', u'z'],
        imports=[
            FOO_TABLE,
        ],
        tokens=[
            _T(u'a', _SID_START + 0, _L(1)),
            _T(u'b', _SID_START + 1, _L(2)),
            _T(u'c', _SID_START + 2, _L(3)),
            _T(u'y', _SID_START + 3),
            _T(u'z', _SID_START + 4),
        ],
    ),
    _P(
        desc='IMPORT DUP SYSTEM',
        symbol_texts=[u'c', u'imports', u'z'],
        imports=[
            COLLIDE_TABLE,
            FOO_TABLE,
        ],
        tokens=[
            _T(u'a', _COLLIDE_SID_START + 0, _L(1)),
            _T(u'b', _COLLIDE_SID_START + 1, _L(2)),
            _T(u'c', _COLLIDE_SID_START + 2, _L(3)),
            _T(u'z', _COLLIDE_SID_START + 3),
        ],
        symbol_count_override=len(COLLIDE_TABLE) + len(FOO_TABLE) + 1
    ),
)
def test_local_symbols(p):
    table = symbols.local_symbol_table(p.imports)
    for text in p.symbol_texts:
        table.intern(text)

    sys_table = symbols.SYSTEM_SYMBOL_TABLE
    if p.symbol_count_override is not None:
        user_symbol_count = p.symbol_count_override
    else:
        user_symbol_count = len(p.tokens)
    assert (user_symbol_count + len(sys_table)) == len(table)

    assert symbols.SYMBOL_ZERO_TOKEN

    for token in itertools.chain(sys_table, p.tokens):
        assert token == table[token.text]
        assert token == table[token.sid]

    with pytest.raises(KeyError):
        table[-1]

    with pytest.raises(KeyError):
        table[u'no_such_symbol']

    with pytest.raises(KeyError):
        table[1000]




# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import statistics

# singleton NormalDist that is the _unit normal distribution_, used for getting z-scores for confidence levels
_unit_normal = statistics.NormalDist()


class SampleDist(list):
    """
    Represents a normal-ish sample distribution. Similar to [NormalDist], but retains the source data.
    """

    def __init__(self, data: list):
        super().__init__(data)
        self.__max = max(data)
        self.__min = min(data)
        self.__mean = statistics.fmean(data)  # This runs faster than the mean() function, and it always returns a float
        self.__stdev = statistics.stdev(data)
        self.__variance = statistics.variance(data, self.__mean)

    @property
    def min(self):
        """Return the min value in the sample set."""
        return self.__min

    @property
    def max(self):
        """Return the max value in the sample set."""
        return self.__max

    @property
    def mean(self):
        """Return the mean of the sample distribution."""
        return self.__mean

    @property
    def variance(self):
        """Return the variance of the sample distribution."""
        return self.__variance

    @property
    # Sometimes also spelled "stddev", but Python Std Lib statistic module uses "stdev".
    def stdev(self):
        """Return the standard deviation of the sample distribution."""
        return self.__stdev

    @property
    def rstdev(self):
        """Return the standard deviation of the sample distribution as a ratio relative to the sample mean"""
        return self.__stdev / self.__mean

    def margin_of_error(self, confidence: float):
        """
        Return the margin of error of the mean of this sample distribution at the given confidence level. Can be
        combined with the mean to express the confidence interval as `mean ± error`.
        """
        z_score = _unit_normal.inv_cdf((1 + confidence) / 2.)
        return z_score * self.stdev / (len(self) ** .5)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from typing import NamedTuple, Sequence

from pytest import raises

from tests import is_exception, listify

from amazon.ion.core import IonEventType
from tests.event_aliases import END
from tests.event_aliases import NEXT


def add_depths(events):
    """Adds the appropriate depths to an iterable of :class:`IonEvent`."""
    depth = 0
    for event in events:
        if is_exception(event):
            # Special case for expectation of exception.
            yield event
        else:
            if event.event_type == IonEventType.CONTAINER_END:
                depth -= 1

            if event.event_type.is_stream_signal:
                yield event
            else:
                yield event.derive_depth(depth)

            if event.event_type == IonEventType.CONTAINER_START:
                depth += 1


class ReaderParameter(NamedTuple):
    desc: str
    event_pairs: Sequence
    is_unicode: bool = False

    def __str__(self):
        return self.desc


def reader_scaffold(reader, event_pairs):
    input_events = (e for e, _ in event_pairs)
    output_events = add_depths(e for _, e in event_pairs)
    for read_event, expected in zip(input_events, output_events):
        if is_exception(expected):
            with raises(expected):
                reader.send(read_event).value  # Forces evaluation of all value thunks.
        else:
            actual = reader.send(read_event)
            assert expected == actual


def value_iter(event_func, values, *args):
    """Generates input/output event pairs from a sequence whose first element is the raw data and the following
    elements are the expected output events.
    """
    for seq in values:
        data = seq[0]
        event_pairs = list(event_func(data, seq[1:], *args))
        yield data, event_pairs


def all_top_level_as_one_stream_params(iterator, *args):
    @listify
    def generate_event_pairs():
        yield (NEXT, END)
        for data, event_pairs in iterator(*args):
            for event_pair in event_pairs:
                yield event_pair
            yield (NEXT, END)

    yield ReaderParameter(
        desc='TOP LEVEL ALL',
        event_pairs=generate_event_pairs()
    )



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import sys
import traceback

_IMPORT_ALL = None
try:
    from amazon.ion import *
except:
    traceback.print_exc()
    _IMPORT_ALL = sys.exc_info()


def test_import_all():
    assert _IMPORT_ALL is None



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from amazon.ion.core import IonType
from amazon.ion.simple_types import IonPyList, IonPyDict, IonPyNull, IonPyBool, IonPyInt, IonPyFloat, IonPyDecimal, \
    IonPyTimestamp, IonPyText, IonPyBytes, IonPySymbol
from amazon.ion.symbols import SymbolToken
from amazon.ion.simpleion import dumps, loads
from base64 import b64encode
import datetime
from decimal import Decimal
import json
import pytest
import sys

is_pypy = hasattr(sys, "pypy_version_info")

if is_pypy:
    # PyPy is not supported. Expect an ImportError.
    with pytest.raises(ImportError):
        from amazon.ion.json_encoder import IonToJSONEncoder
else:
    from amazon.ion.json_encoder import IonToJSONEncoder


def test_null():
    if is_pypy:
        return

    ion_types = [
        (IonPyNull, IonType.NULL),
        (IonPyBool, IonType.BOOL),
        (IonPyInt, IonType.INT),
        (IonPyFloat, IonType.FLOAT),
        (IonPyDecimal, IonType.DECIMAL),
        (IonPyTimestamp, IonType.TIMESTAMP),
        (IonPyText, IonType.STRING),
        (IonPySymbol, IonType.SYMBOL),
        (IonPyBytes, IonType.BLOB),
        (IonPyBytes, IonType.CLOB),
        (IonPyDict, IonType.STRUCT),
        (IonPyList, IonType.LIST),
        (IonPyList, IonType.SEXP)
    ]
    for ion_class, ion_type in ion_types:
        ion_value = ion_class.from_value(ion_type, None)
        json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
        assert json_string == 'null'


def test_bool():
    if is_pypy:
        return

    ion_value = loads(dumps(False))
    assert isinstance(ion_value, IonPyBool) and ion_value.ion_type == IonType.BOOL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == 'false'


def test_int():
    if is_pypy:
        return

    ion_value = loads(dumps(-123))
    assert isinstance(ion_value, IonPyInt) and ion_value.ion_type == IonType.INT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '-123'


def test_float():
    if is_pypy:
        return

    ion_value = loads(dumps(float(123.456)))
    assert isinstance(ion_value, IonPyFloat) and ion_value.ion_type == IonType.FLOAT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '123.456'


def test_float_nan():
    if is_pypy:
        return

    ion_value = loads(dumps(float("NaN")))
    assert isinstance(ion_value, IonPyFloat) and ion_value.ion_type == IonType.FLOAT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == 'null'


def test_float_inf():
    if is_pypy:
        return

    ion_value = loads(dumps(float("Inf")))
    assert isinstance(ion_value, IonPyFloat) and ion_value.ion_type == IonType.FLOAT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == 'null'


def test_decimal():
    if is_pypy:
        return

    ion_value = loads(dumps(Decimal('123.456')))
    assert isinstance(ion_value, IonPyDecimal) and ion_value.ion_type == IonType.DECIMAL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '123.456'


def test_decimal_exp():
    if is_pypy:
        return

    ion_value = loads(dumps(Decimal('1.23456e2')))
    assert isinstance(ion_value, IonPyDecimal) and ion_value.ion_type == IonType.DECIMAL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '123.456'


def test_decimal_exp_negative():
    if is_pypy:
        return

    ion_value = loads(dumps(Decimal('12345.6e-2')))
    assert isinstance(ion_value, IonPyDecimal) and ion_value.ion_type == IonType.DECIMAL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '123.456'


def test_decimal_exp_large():
    if is_pypy:
        return

    ion_value = loads(dumps(Decimal('123.456e34')))
    assert isinstance(ion_value, IonPyDecimal) and ion_value.ion_type == IonType.DECIMAL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '1.23456e+36'


def test_decimal_exp_large_negative():
    if is_pypy:
        return

    ion_value = loads(dumps(Decimal('123.456e-34')))
    assert isinstance(ion_value, IonPyDecimal) and ion_value.ion_type == IonType.DECIMAL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '1.23456e-32'


def test_timestamp():
    if is_pypy:
        return

    ion_value = loads(dumps(datetime.datetime(2010, 6, 15, 3, 30, 45)))
    assert isinstance(ion_value, IonPyTimestamp) and ion_value.ion_type == IonType.TIMESTAMP
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '"2010-06-15 03:30:45"'


def test_symbol():
    if is_pypy:
        return

    ion_value = loads(dumps(SymbolToken(str("Symbol"), None)))
    assert isinstance(ion_value, IonPySymbol) and ion_value.ion_type == IonType.SYMBOL
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '"Symbol"'


def test_string():
    if is_pypy:
        return

    ion_value = loads(dumps(str("String")))
    assert isinstance(ion_value, IonPyText) and ion_value.ion_type == IonType.STRING
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '"String"'


def test_clob():
    if is_pypy:
        return

    ion_value = loads(dumps(IonPyBytes.from_value(IonType.CLOB, bytearray.fromhex("06 49 6f 6e 06"))))
    assert isinstance(ion_value, IonPyBytes) and ion_value.ion_type == IonType.CLOB
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '"\\u0006Ion\\u0006"'


def test_blob():
    if is_pypy:
        return

    ion_value = loads(dumps(bytes("Ion", "ASCII")))
    assert isinstance(ion_value, IonPyBytes) and ion_value.ion_type == IonType.BLOB
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '"SW9u"'


def test_list():
    if is_pypy:
        return

    ion_value = loads(dumps([str("Ion"), 123]))
    assert isinstance(ion_value, IonPyList) and ion_value.ion_type == IonType.LIST
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '["Ion", 123]'


def test_sexp():
    if is_pypy:
        return

    value = (str("Ion"), 123)
    ion_value = loads(dumps(loads(dumps(value, tuple_as_sexp=True))))
    assert isinstance(ion_value, IonPyList) and ion_value.ion_type == IonType.SEXP
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '["Ion", 123]'


def test_struct():
    if is_pypy:
        return

    value = {
        str("string_value"): str("Ion"),
        str("int_value"): 123,
        str("nested_struct"): {
            str("nested_value"): str("Nested Ion")
        }
    }
    ion_value = loads(dumps(value))
    assert isinstance(ion_value, IonPyDict) and ion_value.ion_type == IonType.STRUCT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    expected_string = '{"string_value": "Ion", "int_value": 123, "nested_struct": {"nested_value": "Nested Ion"}}'
    if not json_string == expected_string:
        # Assert as objects to handle different Python versions' JSON string key ordering
        assert json.loads(json_string) == json.loads(expected_string)


def test_annotation_suppression():
    if is_pypy:
        return

    ion_value = loads(dumps(IonPyInt.from_value(IonType.INT, 123, str("Annotation"))))
    assert isinstance(ion_value, IonPyInt) and ion_value.ion_type == IonType.INT
    json_string = json.dumps(ion_value, cls=IonToJSONEncoder)
    assert json_string == '123'



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Methods for writing binary Ion Int, VarInt, UInt, and VarUInt fields.

Whenever the following words are used in variable or method names, their meaning is as
defined below.

- ``signed``: applies only to Int and VarInt - the signed fields (sign bit in first octet).
- ``unsigned``: applies only to UInt and VarUInt - the unsigned fields (no sign bit).
- ``variable``: applies only to VarInt and VarUInt - the variable-length fields (end bit required).
- ``fixed``: applies only to Int and UInt - the fixed-length fields (no end bit).
- ``Int``: applies only to the Int field - the signed, fixed-length field.
- ``varint``: applies only to the VarInt field - the signed, variable-length field.
- ``uint``: applies only to the UInt field - the unsigned, fixed-length field.
- ``varuint``: applies only to the VarUInt field - the unsigned, variable-length field.
"""

from .util import bit_length

_VARIABLE_END_BIT_MASK = 0b10000000  # Both VarInt and VarUInt

_INT_SIGN_BIT_MASK = 0b10000000  # fixed Int only
_VARINT_SIGN_BIT_MASK = 0b01000000  # VarInt only

_VARIABLE_BITS_PER_OCTET = 7  # Both VarInt and VarUInt (has end bit) (exclusive of first octet)
_FIXED_BITS_PER_OCTET = 8  # Both fixed Int and fixed UInt (exclusive of first octet)

_OCTET_MASKS = {
    _VARIABLE_BITS_PER_OCTET: 0x7F,
    _FIXED_BITS_PER_OCTET: 0xFF,
}


def _write_varint(buf, value):
    """Writes the given integer value into the given buffer as a binary Ion VarInt field.

    Args:
        buf (Sequence): The buffer into which the VarInt will be written, in the form of
            integer octets.
        value (int): The value to write as a VarInt.

    Returns:
        int: The number of octets written.
    """
    return _write_signed(buf, value, _field_cache.get_varint, _write_varint_uncached)


def _write_varint_uncached(buf, value):
    return _write_signed_uncached(buf, value, _VARINT_SIGN_BIT_MASK, _VARIABLE_BITS_PER_OCTET,
                                  _VARIABLE_END_BIT_MASK)


def _write_int(buf, value):
    """Writes the given integer value into the given buffer as a binary Ion Int field.

    Args:
        buf (Sequence): The buffer into which the Int will be written,
            in the form of integer octets.
        value (int): The value to write as a Int.

    Returns:
        int: The number of octets written.
    """
    return _write_signed(buf, value, _field_cache.get_int, _write_int_uncached)


def _write_int_uncached(buf, value):
    return _write_signed_uncached(buf, value, _INT_SIGN_BIT_MASK, _FIXED_BITS_PER_OCTET)


def _write_signed(buf, value, cached_func, uncached_func):
    if _field_cache.SIGNED_MIN <= value < _field_cache.SIGNED_MAX:
        buf.append(cached_func(value))
        return 1
    return uncached_func(buf, value)


def _write_signed_uncached(buf, value, sign_bit_mask, bits_per_octet, end_bit=0):
    magnitude = value
    sign_bit = 0
    if value < 0:
        magnitude = -magnitude
        sign_bit = sign_bit_mask
    return _write_base(buf, magnitude, bits_per_octet, end_bit, sign_bit, is_signed=True)


def _write_varuint(buf, value):
    """Writes the given integer value into the given buffer as a binary Ion VarUInt.

    Args:
        buf (Sequence): The buffer into which the VarUInt will be written,
            in the form of integer octets.
        value (int): The value to write as a VarUInt.

    Returns:
        int: The number of octets written.
    """
    return _write_unsigned(buf, value, _field_cache.get_varuint, _write_varuint_uncached)


def _write_varuint_uncached(buf, value):
    return _write_base(buf, value, _VARIABLE_BITS_PER_OCTET, _VARIABLE_END_BIT_MASK)


def _write_uint(buf, value):
    """Writes the given integer value into the given buffer as a binary Ion UInt.

    Args:
        buf (Sequence): The buffer into which the UInt will be written,
            in the form of integer octets.
        value (int): The value to write as a UInt.

    Returns:
        int: The number of octets written.
    """
    return _write_unsigned(buf, value, _field_cache.get_uint, _write_uint_uncached)


def _write_uint_uncached(buf, value):
    return _write_base(buf, value, _FIXED_BITS_PER_OCTET)


def _write_unsigned(buf, value, cached_func, uncached_func):
    if value < _CACHE_SIZE:
        buf.append(cached_func(value))
        return 1
    return uncached_func(buf, value)


def _write_base(buf, value, bits_per_octet, end_bit=0, sign_bit=0, is_signed=False):
    """Write a field to the provided buffer.

    Args:
        buf (Sequence): The buffer into which the UInt will be written
            in the form of integer octets.
        value (int): The value to write as a UInt.
        bits_per_octet (int): The number of value bits (i.e. exclusive of the end bit, but
            inclusive of the sign bit, if applicable) per octet.
        end_bit (Optional[int]): The end bit mask.
        sign_bit (Optional[int]): The sign bit mask.

    Returns:
        int: The number of octets written.
    """
    if value == 0:
        buf.append(sign_bit | end_bit)
        return 1
    num_bits = bit_length(value)
    num_octets = num_bits // bits_per_octet
    # 'remainder' is the number of value bits in the first octet.
    remainder = num_bits % bits_per_octet
    if remainder != 0 or is_signed:
        # If signed, the first octet has one fewer bit available, requiring another octet.
        num_octets += 1
    else:
        # This ensures that unsigned values that fit exactly are not shifted too far.
        remainder = bits_per_octet
    for i in range(num_octets):
        octet = 0
        if i == 0:
            octet |= sign_bit
        if i == num_octets - 1:
            octet |= end_bit
        # 'remainder' is used for alignment such that only the first octet
        # may contain insignificant zeros.
        octet |= ((value >> (num_bits - (remainder + bits_per_octet * i))) & _OCTET_MASKS[bits_per_octet])
        buf.append(octet)
    return num_octets


_CACHE_SIZE = 64


class _FieldCache:
    """Contains caches for small Int, UInt, VarInt, and VarUInt values.

    For unsigned fields, values between 0 and 63 will have their representations cached.
    For signed fields, values between -32 and 31 will have their representations cached.
    It is likely that a large proportion of subfields fit in these ranges. For example:
    - All values with lengths less than 64 bytes will retrieve their length
        subfield from the cache.
    - All integer values with magnitudes less than 64 will retrieve their 'magnitude'
        subfields from the cache.
    - Decimal values with exponents and/or coefficients within the signed cache range
        will retrieve those fields from the cache.
    - All timestamps will retrieve their month, day, hour, minute, and second fields
        from the cache (otherwise it would be an invalid datetime). The fractional fields
        may be cached if they meet the decimal value requirement from the previous bullet.
    - The first 64 symbols will have their symbol ID fields retrieved from the cache.
        This applies to symbol values, field names, and annotations.
    - If a value has less than 64 bytes of annotations (hopefully this is true for all
        values), then its annotation wrapper's 'annotation length' field will be cached.
    """
    @classmethod
    def _signed_value(cls, index):
        return index - cls.__HALF

    @classmethod
    def _signed_index(cls, value):
        return cls.__HALF + value

    __HALF = _CACHE_SIZE // 2

    def __init__(self):
        self.SIZE = _CACHE_SIZE
        self.SIGNED_MAX = _FieldCache._signed_value(self.SIZE)
        self.SIGNED_MIN = _FieldCache._signed_value(0)
        self._cached_ints = bytearray()
        self._cached_varints = bytearray()
        self._cached_uints = bytearray()
        self._cached_varuints = bytearray()
        self._fill_cache()

    def _fill_cache(self):
        # Each cached_value is an index, but since the caches are self-resizing sequences,
        # they are not directly used that way here. Each _write* method below appends to
        # the given cache. Since each cache starts empty, the field value ends up at
        # index 'cached_value' in the cache. For unsigned fields, the cached field represents
        # cached_value directly. In other words, the UInt cache contains the UInt field
        # representation of '32' at index '32'. For signed fields, the cached field represents
        # (cached_value - 32) to give a selection of positive and negative values.
        # In other words, the Int cache contains the Int field representation of '0' at index '32'.
        #
        # The size of 64 was chosen because it is the maximum magnitude that fits in a
        # 1-byte VarInt field (the smallest max magnitude of any 1-byte field representation).
        # Therefore, each value of 'cached_value' is guaranteed to fit in one byte for any
        # of the four field types (this is verified by the assertion).
        # Note that different sizes could be used for each of the caches,
        # but for simplicity, 64 is used for all.
        for cached_value in range(self.SIZE):
            signed_value = _FieldCache._signed_value(cached_value)
            assert _write_int_uncached(self._cached_ints, signed_value) == 1
            assert _write_varint_uncached(self._cached_varints, signed_value) == 1
            assert _write_uint_uncached(self._cached_uints, cached_value) == 1
            assert _write_varuint_uncached(self._cached_varuints, cached_value) == 1

    def get_int(self, value):
        return self._cached_ints[_FieldCache._signed_index(value)]

    def get_varint(self, value):
        return self._cached_varints[_FieldCache._signed_index(value)]

    def get_uint(self, value):
        return self._cached_uints[value]

    def get_varuint(self, value):
        return self._cached_varuints[value]

_field_cache = _FieldCache()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from base64 import b64encode
from datetime import datetime, timedelta

from decimal import Decimal

import sys
from typing import NamedTuple, Sequence, Any

from pytest import raises

from amazon.ion.core import ION_STREAM_END_EVENT, IonEventType, IonEvent, IonType, timestamp, OffsetTZInfo, \
    TimestampPrecision
from amazon.ion.symbols import SYMBOL_ZERO_TOKEN, SymbolToken
from amazon.ion.writer import WriteEventType
from tests import is_exception
from tests import noop_manager

_STREAM_END_EVENT = (ION_STREAM_END_EVENT,)

ION_ENCODED_INT_ZERO = 0x20
VARUINT_END_BYTE = 0x80

_D = Decimal
_DT = datetime
_IT = IonType


class WriterParameter(NamedTuple):
    desc: str
    events: Sequence[IonEvent]
    expected: Any

    def __str__(self):
        return self.desc


def _scalar_p(ion_type, value, expected, force_stream_end):
    events = (IonEvent(IonEventType.SCALAR, ion_type, value),)
    if force_stream_end:
        events += _STREAM_END_EVENT
    return WriterParameter(
        desc='SCALAR %s - %r - %r' % (ion_type.name, value, expected),
        events=events,
        expected=expected,
    )


def _convert_symbol_pairs_to_string_pairs(symbol_pairs):
    for value, literals in symbol_pairs:
        final_literals = ()
        if not isinstance(literals, (tuple, list)):
            literals = (literals,)
        for literal in literals:
            if literal.decode('utf-8')[0] == "'":
                final_literals += (literal.replace(b"'", b'"'),)
            else:
                # Add quotes to unquoted symbols
                final_literals += ((b'"' + literal + b'"'),)
        yield value, final_literals


def _convert_clob_pairs(clob_pairs):
    for value, literal in clob_pairs:
        yield (value, b'{{' + b64encode(value) + b'}}')


_SIMPLE_SYMBOLS_TEXT=(
    (u'', br"''"),
    (u'\u0000', (br"'\x00'", br"'\0'")),
    (u'4hello', br"'4hello'"),
    (u'hello', br"hello"),
    (u'_hello_world', br"_hello_world"),
    (u'null', br"'null'"),
    (u'hello world', br"'hello world'"),
    (u'hello\u0009\x0a\x0dworld', br"'hello\t\n\rworld'"),
    (u'hello\aworld', (br"'hello\x07world'", br"'hello\aworld'")),
    (u'hello\u3000world', (br"'hello\u3000world'", b"'hello\xe3\x80\x80world'")),  # A full width space.
    (u'hello\U0001f4a9world', (br"'hello\U0001f4a9world'", b"'hello\xf0\x9f\x92\xa9world'")),  # A 'pile of poo' emoji code point.
)
_SIMPLE_STRINGS_TEXT=tuple(_convert_symbol_pairs_to_string_pairs(_SIMPLE_SYMBOLS_TEXT))

_SIMPLE_CLOBS_TEXT=(
    (b'', br'{{""}}'),
    (b'\x00', (br'{{"\x00"}}', br'{{"\0"}}')),
    (b'hello', br'{{"hello"}}'),
    (b'hello\x09\x0a\x0dworld', br'{{"hello\t\n\rworld"}}'),
    (b'hello\x7Eworld', br'{{"hello~world"}}'),
    (b'hello\xFFworld', (br'{{"hello\xffworld"}}', br'{{"hello\xFFworld"}}')),
)
_SIMPLE_BLOBS_TEXT=tuple(_convert_clob_pairs(_SIMPLE_CLOBS_TEXT))

if sys.version_info < (2, 7):
    # Python < 2.7 was not good at some float irrationals.
    _FLOAT_1_1_ENC = b'1.1000000000000001e0'
    _FLOAT_2_E_NEG_15_ENC = b'2.0000000000000002e-15'
else:
    _FLOAT_1_1_ENC = b'1.1e0'
    _FLOAT_2_E_NEG_15_ENC = b'2e-15'

SIMPLE_SCALARS_MAP_TEXT = {
    _IT.NULL: (
        (None, b'null'),
    ),
    _IT.BOOL: (
        (None, b'null.bool'),
        (True, b'true'),
        (False, b'false'),
    ),
    _IT.INT: (
        (None, b'null.int'),
        (-1, b'-1'),
        (0, b'0'),
        (1, b'1'),
        (0xFFFFFFFF, b'4294967295'),
        (-0xFFFFFFFF, b'-4294967295'),
        (0xFFFFFFFFFFFFFFFF, b'18446744073709551615'),
        (-0xFFFFFFFFFFFFFFFF, b'-18446744073709551615'),
    ),
    _IT.FLOAT: (
        (None, b'null.float'),
        (float('NaN'), b'nan'),
        (float('+Inf'), b'+inf'),
        (float('-Inf'), b'-inf'),
        (-0.0, (b'-0.0e0', b'-0e0')),
        (0.0, (b'0.0e0', b'0e0')),
        (1.0, (b'1.0e0', b'1e+0')),
        (-9007199254740991.0, (b'-9007199254740991.0e0', b'-9007199254740991e+0')),
        (2.0e-15, (_FLOAT_2_E_NEG_15_ENC, b'2.0000000000000001554e-15')),
        (1.1, (_FLOAT_1_1_ENC, b'1.1000000000000000888e+0')),
        (1.1999999999999999555910790149937383830547332763671875e0, (b'1.2e0', b'1.1999999999999999556e+0')),
    ),
    _IT.DECIMAL: (
        (None, b'null.decimal'),
        (_D('-0.0'), b'-0.0'),
        (_D('0'), b'0d0'),
        (_D('0e100'), b'0d+100'),
        (_D('0e-15'), b'0d-15'),
        (_D('-1e1000'), b'-1d+1000'),
        (_D('-4.412111311414141e1000'), b'-4.412111311414141d+1000'),
        # TODO C extension doesn't support this much decimal digits. comment it for now.
        # (_D('1.1999999999999999555910790149937383830547332763671875e0'),
        #     b'1.1999999999999999555910790149937383830547332763671875'),
    ),
    _IT.TIMESTAMP: (
        (None, b'null.timestamp'),
        (_DT(2016, 1, 1), b'2016-01-01T00:00:00.000000-00:00'),
        (_DT(2016, 1, 1, 12), b'2016-01-01T12:00:00.000000-00:00'),
        (_DT(2016, 1, 1, 12, 34, 12), b'2016-01-01T12:34:12.000000-00:00'),
        (_DT(2016, 1, 1, 12, 34, 12, 555000), b'2016-01-01T12:34:12.555000-00:00'),
        (_DT(2016, 1, 1, 12, 34, 12, tzinfo=OffsetTZInfo()), b'2016-01-01T12:34:12.000000Z'),
        (_DT(2016, 1, 1, 12, 34, 12, tzinfo=OffsetTZInfo(timedelta(hours=-7))),
            b'2016-01-01T12:34:12.000000-07:00'),
        (timestamp(year=1, month=1, day=1, precision=TimestampPrecision.DAY), (b'0001-01-01T',  b'0001-01-01')),
        (timestamp(year=1, month=1, day=1, off_minutes=-1, precision=TimestampPrecision.SECOND),
         b'0001-01-01T00:00:00-00:01'),
        (
            timestamp(year=1, month=1, day=1, hour=0, minute=0, second=0,
                      microsecond=1, precision=TimestampPrecision.SECOND),
            b'0001-01-01T00:00:00.000001-00:00'
        ),
        (
            timestamp(year=1, month=1, day=1, hour=0, minute=0, second=0,
                      microsecond=100000, precision=TimestampPrecision.SECOND, fractional_precision=1),
            b'0001-01-01T00:00:00.1-00:00'
        ),
        (timestamp(2016, precision=TimestampPrecision.YEAR), b'2016T'),
        (timestamp(2016, off_hours=0, precision=TimestampPrecision.YEAR), b'2016T'),
        (
            timestamp(2016, 2, 1, 0, 1, off_minutes=1, precision=TimestampPrecision.MONTH),
            b'2016-02T'
        ),
        (
            timestamp(2016, 2, 1, 23, 0, off_hours=-1, precision=TimestampPrecision.DAY),
            (b'2016-02-01T', b'2016-02-01')
        ),
        (
            timestamp(2016, 2, 2, 0, 0, off_hours=-7, precision=TimestampPrecision.MINUTE),
            b'2016-02-02T00:00-07:00'
        ),
        (
           timestamp(2016, 2, 2, 0, 0, 30, off_hours=-7, precision=TimestampPrecision.SECOND),
           b'2016-02-02T00:00:30-07:00'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 1000, off_hours=-7,
                      precision=TimestampPrecision.SECOND),
            # When fractional_precision not specified, defaults to 6 (same as regular datetime).
            b'2016-02-02T00:00:30.001000-07:00'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 1000, off_hours=-7,
                      precision=TimestampPrecision.SECOND, fractional_precision=3),
            b'2016-02-02T00:00:30.001-07:00'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 100000, off_hours=-7,
                      precision=TimestampPrecision.SECOND, fractional_precision=1),
            b'2016-02-02T00:00:30.1-07:00'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, precision=TimestampPrecision.SECOND,
                      fractional_seconds=Decimal('0.000010000')), b'2016-02-02T00:00:30.000010000-00:00'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, precision=TimestampPrecision.SECOND,
                      fractional_seconds=Decimal('0.7e-500')),
            (b'2016-02-02T00:00:30.' + b'0' * 500 + b'7-00:00', b'2016-02-02T00:00:30.000000000-00:00')
        )
    ),
    _IT.SYMBOL: (
        (None, b'null.symbol'),
        (SymbolToken(None, 4), (b'$4', b'name')),  # System symbol 'name'.
        (SymbolToken(u'a token', 400), b"'a token'"),
    ) + _SIMPLE_SYMBOLS_TEXT,
    _IT.STRING: (
        (None, b'null.string'),
    ) + _SIMPLE_STRINGS_TEXT,
    _IT.CLOB: (
        (None, b'null.clob'),
    ) + _SIMPLE_CLOBS_TEXT,
    _IT.BLOB: (
        (None, b'null.blob'),
    ) + _SIMPLE_BLOBS_TEXT,
    _IT.LIST: (
        (None, b'null.list'),
    ),
    _IT.SEXP: (
        (None, b'null.sexp'),
    ),
    _IT.STRUCT: (
        (None, b'null.struct'),
    ),
}

SIMPLE_SCALARS_MAP_BINARY = {
    _IT.NULL: (
        (None, b'\x0F'),
    ),
    _IT.BOOL: (
        (None, b'\x1F'),
        (False, b'\x10'),
        (True, b'\x11')
    ),
    _IT.INT: (
        (None, b'\x2F'),
        (0, b'\x20'),
        (1, b'\x21\x01'),
        (-1, b'\x31\x01'),
        (0xFFFFFFFF, b'\x24\xFF\xFF\xFF\xFF'),
        (-0xFFFFFFFF, b'\x34\xFF\xFF\xFF\xFF'),
        (0xFFFFFFFFFFFFFFFF, b'\x28\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'),
        (-0xFFFFFFFFFFFFFFFF, b'\x38\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'),
        (0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
         b'\x2E\x90\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'),
        (-0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF,
         b'\x3E\x90\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF'),
    ),
    _IT.FLOAT: (
        (None, b'\x4F'),
        (0e0, b'\x40'),
        (1e0, b'\x48\x3F\xF0\x00\x00\x00\x00\x00\x00'),
        (-1e0, b'\x48\xBF\xF0\x00\x00\x00\x00\x00\x00'),
        (1e1, b'\x48\x40\x24\x00\x00\x00\x00\x00\x00'),
        (-1e1, b'\x48\xC0\x24\x00\x00\x00\x00\x00\x00'),
        (float('inf'), b'\x48\x7F\xF0\x00\x00\x00\x00\x00\x00'),
        (float('-inf'), b'\x48\xFF\xF0\x00\x00\x00\x00\x00\x00'),
    ),
    _IT.DECIMAL: (
        (None, b'\x5F'),
        (_D(0), b'\x50'),
        (_D(0).copy_negate(), b'\x52\x80\x80'),
        (_D("1e1"), b'\x52\x81\x01'),
        (_D("1e0"), b'\x52\x80\x01'),
        (_D("1e-1"), b'\x52\xC1\x01'),
        (_D("0e-1"), b'\x51\xC1'),
        (_D("0e1"), b'\x51\x81'),
        (_D("-1e1"), b'\x52\x81\x81'),
        (_D("-1e0"), b'\x52\x80\x81'),
        (_D("-1e-1"), b'\x52\xC1\x81'),
        (_D("-0e-1"), b'\x52\xC1\x80'),
        (_D("-0e1"), b'\x52\x81\x80'),
    ),
    _IT.TIMESTAMP: (
        (None, b'\x6F'),
        # TODO Clarify whether there's a valid zero-length Timestamp representation.
        (_DT(year=1, month=1, day=1), b'\x68\xC0\x81\x81\x81\x80\x80\x80\xc6'),
        (_DT(year=1, month=1, day=1, tzinfo=OffsetTZInfo(timedelta(minutes=-1))),
         b'\x68\xC1\x81\x81\x81\x80\x81\x80\xc6'),
        (_DT(year=1, month=1, day=1, hour=0, minute=0, second=0, microsecond=1),
         b'\x69\xC0\x81\x81\x81\x80\x80\x80\xC6\x01'),
        (timestamp(year=1, month=1, day=1, precision=TimestampPrecision.DAY), b'\x64\xC0\x81\x81\x81'),
        (timestamp(year=1, month=1, day=1, off_minutes=-1, precision=TimestampPrecision.SECOND),
         b'\x67\xC1\x81\x81\x81\x80\x81\x80'),
        (
            timestamp(year=1, month=1, day=1, hour=0, minute=0, second=0,
                      microsecond=1, precision=TimestampPrecision.SECOND),
            b'\x69\xC0\x81\x81\x81\x80\x80\x80\xC6\x01'
        ),
        (
            timestamp(year=1, month=1, day=1, hour=0, minute=0, second=0,
                      microsecond=100000, precision=TimestampPrecision.SECOND, fractional_precision=1),
            b'\x69\xC0\x81\x81\x81\x80\x80\x80\xC1\x01'
        ),
        (timestamp(2016, precision=TimestampPrecision.YEAR), b'\x63\xC0\x0F\xE0'),  # -00:00
        (timestamp(2016, off_hours=0, precision=TimestampPrecision.YEAR),
            (b'\x63\x80\x0F\xE0', b'\x63\xC0\x0F\xE0')),
        (
            timestamp(2016, 2, 1, 0, 1, off_minutes=1, precision=TimestampPrecision.MONTH),
            (b'\x64\x81\x0F\xE0\x82', b'\x64\xC0\x0F\xE0\x82')
        ),
        (
            timestamp(2016, 2, 1, 23, 0, off_hours=-1, precision=TimestampPrecision.DAY),
            (b'\x65\xFC\x0F\xE0\x82\x82', b'\x65\xC0\x0F\xE0\x82\x81')
        ),
        (
            timestamp(2016, 2, 2, 0, 0, off_hours=-7, precision=TimestampPrecision.MINUTE),
            b'\x68\x43\xA4\x0F\xE0\x82\x82\x87\x80'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, off_hours=-7, precision=TimestampPrecision.SECOND),
            b'\x69\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 1000, off_hours=-7,
                      precision=TimestampPrecision.SECOND),
            # When fractional_precision not specified, defaults to 6 (same as regular datetime).
            b'\x6C\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC6\x03\xE8'  # The last three octets represent 1000d-6
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 1000, off_hours=-7,
                      precision=TimestampPrecision.SECOND, fractional_precision=3),
            b'\x6B\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC3\x01'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, 100000, off_hours=-7,
                      precision=TimestampPrecision.SECOND, fractional_precision=1),
            b'\x6B\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC1\x01'
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, precision=TimestampPrecision.SECOND,
                      fractional_seconds=Decimal('0.000010000')),
            (b'\x6B\xC0\x0F\xE0\x82\x82\x80\x80\x9E\xC9\x27\x10',
             b'\x6A\xC0\x0F\xE0\x82\x82\x80\x80\x9E\xC6\x0A')
        ),
        (
            timestamp(2016, 2, 2, 0, 0, 30, precision=TimestampPrecision.SECOND,
                      fractional_seconds=Decimal('0.7e-500')),
            (b'\x6B\xC0\x0F\xE0\x82\x82\x80\x80\x9E\x43\xF5\x07',
             b'\x69\xC0\x0F\xE0\x82\x82\x80\x80\x9E\xC9')
        )
    ),
    _IT.SYMBOL: (
        (None, b'\x7F'),
        (SYMBOL_ZERO_TOKEN, b'\x70'),
        (SymbolToken(u'$ion', 1), b'\x71\x01'),
    ),
    _IT.STRING: (
        (None, b'\x8F'),
        (u'', b'\x80'),
        (u'abc', b'\x83abc'),
        (u'abcdefghijklmno', b'\x8E\x8Fabcdefghijklmno'),
        (u'a\U0001f4a9c', b'\x86' + bytearray([b for b in u'a\U0001f4a9c'.encode('utf-8')])),
        (u'a\u0009\x0a\x0dc', b'\x85' + bytearray([b for b in 'a\t\n\rc'.encode('utf-8')])),
    ),
    _IT.CLOB: (
        (None, b'\x9F'),
        (b'', b'\x90'),
        (b'abc', b'\x93' + b'abc'),
        (b'abcdefghijklmno', b'\x9E\x8Fabcdefghijklmno'),
    ),
    _IT.BLOB: (
        (None, b'\xAF'),
        (b'', b'\xA0'),
        (b'abc', b'\xA3' + b'abc'),
        (b'abcdefghijklmno', b'\xAE\x8Fabcdefghijklmno'),
    ),
    _IT.LIST: (
        (None, b'\xBF'),
    ),
    _IT.SEXP: (
        (None, b'\xCF'),
    ),
    _IT.STRUCT: (
        (None, b'\xDF'),
    ),
}


def generate_scalars(scalars_map, force_stream_end=False):
    for ion_type, values in iter(scalars_map.items()):
        for native, expected in values:
            yield _scalar_p(ion_type, native, expected, force_stream_end)


def generate_containers(containers_map, force_stream_end=False):
    for ion_type, container in iter(containers_map.items()):
        for container_value_events, expected in container:
            start_event = IonEvent(IonEventType.CONTAINER_START, ion_type)
            end_event = IonEvent(IonEventType.CONTAINER_END, ion_type)
            events = (start_event,) + container_value_events + (end_event,)
            if force_stream_end:
                events += _STREAM_END_EVENT
            yield WriterParameter(
                desc='CONTAINER %s - %r' % (ion_type.name, expected),
                events=events,
                expected=expected,
            )


def assert_writer_events(p, new_writer):
    buf, buf_writer = new_writer()

    ctx = noop_manager()
    if is_exception(p.expected):
        ctx = raises(p.expected)

    result_type = None
    with ctx:
        for event in p.events:
            result_type = buf_writer.send(event)

    if not is_exception(p.expected):
        assert result_type is WriteEventType.COMPLETE

        if isinstance(p.expected, (tuple, list)):
            expecteds = p.expected
        else:
            expecteds = (p.expected,)
        assert_res = False
        for expected in expecteds:
            if expected == buf.getvalue():
                assert_res = True
                break
        assert assert_res



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from google.protobuf import message as _message
from google.protobuf import message_factory as _message_factory
from google.protobuf import reflection as _reflection

from self_describing_proto_pb2 import SelfDescribingMessage

import six


class SelfDescribingProtoSerde:
    """
    This class provides functions for reading and writing self-describing protocol buffers.

    This class uses the technique described in https://protobuf.dev/programming-guides/techniques/#self-description to
    create a mostly self-describing protocol buffer. (It's not truly self-describing because the reader must know that
    the serialized data uses the Self-describing Messages convention.)

    To create a self-describing protocol buffer, see proto_tools.py.

    This class makes no guarantees of thread-safety because each instance of this class has its own single instance of
    the SelfDescribingMessage wrapper that it reuses for every load(s) call.

    TODO: Add support for importing "well-known types" (https://protobuf.dev/reference/protobuf/google.protobuf/).
    """
    def __init__(self, cache_type_info=True, reuse_inner_object=False):
        """
        :param cache_type_info: Controls whether the type descriptor set and generated classes should be cached. Caching
            the generated classes for load(s) can be significantly faster than not caching, but requires more memory for
            the cache. WARNING: there is no cache eviction implemented yet.
            (Has no effect for dump(s).)
        :param reuse_inner_object: Controls whether loads should create a new instance of the inner message class each
            time load(s) is called.
            (Has no effect if cache_type_info is False. Has no effect for dump(s).)
        """
        self._cache_type_info = cache_type_info
        self._reuse_inner_object = reuse_inner_object
        self._cached_outer_object = SelfDescribingMessage()
        self._cached_inner_definitions = {}
        self._cached_inner_objects = {}

    @staticmethod
    def generate_class_definition(type_name, descriptor_set):
        """
        Generate a Python class for the given type_name using the provided descriptor_set.
        """
        messages_types = _message_factory.GetMessages(descriptor_set.file)
        message_type = messages_types[type_name]()

        class DynamicMessage(six.with_metaclass(_reflection.GeneratedProtocolMessageType, _message.Message)):
            DESCRIPTOR = message_type.DESCRIPTOR

        return DynamicMessage

    def _get_inner_object_instance(self, type_name, descriptor_set):
        """
        Get an uninitialized instance of the inner object for the given message.
        """
        if self._cache_type_info:
            descriptor_set_key = descriptor_set.SerializeToString()

            if descriptor_set_key not in self._cached_inner_definitions:
                self._cached_inner_definitions[descriptor_set_key] = {}
            if type_name not in self._cached_inner_definitions[descriptor_set_key]:
                self._cached_inner_definitions[descriptor_set_key][type_name] = \
                    SelfDescribingProtoSerde.generate_class_definition(type_name, descriptor_set)

            clazz = self._cached_inner_definitions[descriptor_set_key][type_name]

            if self._reuse_inner_object:
                if descriptor_set_key not in self._cached_inner_objects:
                    self._cached_inner_objects[descriptor_set_key] = {}
                if type_name not in self._cached_inner_objects[descriptor_set_key]:
                    self._cached_inner_objects[descriptor_set_key][type_name] = clazz()

                return self._cached_inner_objects[descriptor_set_key][type_name]
            else:
                return clazz()
        else:
            clazz = SelfDescribingProtoSerde.generate_class_definition(type_name, descriptor_set)
            return clazz()

    def loads(self, s):
        """
        Deserialize a self-describing protocol buffer from bytes/string.
        """
        outer_obj = self._cached_outer_object
        outer_obj.ParseFromString(s)
        inner_obj = self._get_inner_object_instance(outer_obj.message.type_url, outer_obj.descriptor_set)
        inner_obj.ParseFromString(outer_obj.message.value)
        return inner_obj

    def load(self, fp):
        """
        Deserialize a self-describing protocol buffer from a file.
        """
        return self.loads(fp.read())

    def dumps(self, obj):
        """
        Serialize a protocol buffer message as self-describing protocol buffer bytes/string.
        """
        # It seems to be faster to create a new SelfDescribingMessage for each call than it is to use the cached object
        # and check and/or clear the descriptor for each call.
        outer_object = SelfDescribingMessage()
        obj.DESCRIPTOR.file.CopyToProto(outer_object.descriptor_set.file.add())

        outer_object.message.type_url = obj.DESCRIPTOR.full_name
        outer_object.message.value = obj.SerializeToString()
        return outer_object.SerializeToString()

    def dump(self, obj, fp):
        """
        Serialize a protocol buffer message as a self-describing protocol buffer file.
        """
        fp.write(self.dumps(obj))



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Writer for raw binary Ion values, without symbol table management."""

from datetime import datetime
from decimal import Decimal, localcontext
from enum import IntEnum
from functools import partial

import struct

from amazon.ion.equivalence import _is_float_negative_zero
from amazon.ion.symbols import SymbolToken
from .core import IonEventType, IonType, DataEvent, Transition, TimestampPrecision, TIMESTAMP_FRACTION_PRECISION_FIELD, \
    MICROSECOND_PRECISION, TIMESTAMP_PRECISION_FIELD, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, Timestamp
from .util import coroutine, total_seconds
from .writer import NOOP_WRITER_EVENT, WriteEventType, \
                    writer_trampoline, partial_transition, serialize_scalar, \
                    validate_scalar_value, illegal_state_null
from .writer_binary_raw_fields import _write_varuint, _write_uint, _write_varint, _write_int


class _TypeIds(IntEnum):
    NULL = 0x00
    BOOL_FALSE = 0x10
    BOOL_TRUE = 0x11
    POS_INT = 0x20
    NEG_INT = 0x30
    FLOAT = 0x40
    DECIMAL = 0x50
    TIMESTAMP = 0x60
    SYMBOL = 0x70
    STRING = 0x80
    CLOB = 0x90
    BLOB = 0xA0
    LIST = 0xB0
    SEXP = 0xC0
    STRUCT = 0xD0
    ANNOTATION_WRAPPER = 0xE0


class _Zeros(IntEnum):
    """Single-octet encodings, represented by the type ID and a length nibble of zero.

    Notes:
        Blob, clob, list, and sexp also have single-octet encodings, but the current
        implementation handles these implicitly.
    """
    INT = _TypeIds.POS_INT  # Int zero is always encoded using the positive type ID.
    FLOAT = _TypeIds.FLOAT  # Represents zero.
    DECIMAL = _TypeIds.DECIMAL  # Represents 0d0.
    STRING = _TypeIds.STRING  # Represents the zero-length (empty) string.
    SYMBOL = _TypeIds.SYMBOL  # Represents symbol zero.
    STRUCT = _TypeIds.STRUCT  # Represents a struct with zero fields.

_VARINT_NEG_ZERO = 0xC0  # This refers to the variable-length signed integer subfield.
_INT_NEG_ZERO = 0x80  # This refers to the fixed-length signed integer subfield.

_LENGTH_FLOAT_64 = 0x08
_LENGTH_FIELD_THRESHOLD = 14
_LENGTH_FIELD_INDICATOR = 0x0E
_NULL_INDICATOR = 0x0F


def _null(tid):
    return bytearray([tid | _NULL_INDICATOR])


_NULLS = [
    _null(_TypeIds.NULL),
    _null(_TypeIds.BOOL_FALSE),
    _null(_TypeIds.POS_INT),
    _null(_TypeIds.FLOAT),
    _null(_TypeIds.DECIMAL),
    _null(_TypeIds.TIMESTAMP),
    _null(_TypeIds.SYMBOL),
    _null(_TypeIds.STRING),
    _null(_TypeIds.CLOB),
    _null(_TypeIds.BLOB),
    _null(_TypeIds.LIST),
    _null(_TypeIds.SEXP),
    _null(_TypeIds.STRUCT)
]

_BOOL_TRUE = bytearray([_TypeIds.BOOL_TRUE])
_BOOL_FALSE = bytearray([_TypeIds.BOOL_FALSE])


def _serialize_bool(ion_event):
    if ion_event.value:
        return _BOOL_TRUE
    else:
        return _BOOL_FALSE


def _write_length(buf, length, tid):
    if length < _LENGTH_FIELD_THRESHOLD:
        buf.append(tid | length)
    else:
        buf.append(tid | _LENGTH_FIELD_INDICATOR)
        _write_varuint(buf, length)


def _write_int_value(buf, tid, value):
    value_buf = bytearray()
    length = _write_uint(value_buf, value)
    _write_length(buf, length, tid)
    buf.extend(value_buf)


def _serialize_int(ion_event):
    buf = bytearray()
    value = ion_event.value
    validate_scalar_value(value, int)
    if value == 0:
        buf.append(_Zeros.INT)
    else:
        if value < 0:
            value = -value
            tid = _TypeIds.NEG_INT
        else:
            tid = _TypeIds.POS_INT
        _write_int_value(buf, tid, value)
    return buf


def _serialize_float(ion_event):
    buf = bytearray()
    float_value = ion_event.value
    validate_scalar_value(float_value, float)
    # TODO Assess whether abbreviated encoding of zero is beneficial; it's allowed by spec.
    if float_value.is_integer() and float_value == 0.0 and not _is_float_negative_zero(float_value):
        buf.append(_Zeros.FLOAT)
    else:
        # TODO Add an option for 32-bit representation (length=4) per the spec.
        buf.append(_TypeIds.FLOAT | _LENGTH_FLOAT_64)
        encoded = struct.pack('>d', float_value)
        buf.extend(encoded)
    return buf


def _write_decimal_value(buf, exponent, coefficient, sign=0):
    length = _write_varint(buf, exponent)
    if coefficient:
        # The coefficient is non-zero, so the coefficient field is required.
        length += _write_int(buf, coefficient)
    elif sign:
        # The coefficient is negative zero.
        buf.append(_INT_NEG_ZERO)
        length += 1
    # Else the coefficient is positive zero and the field is omitted.
    return length


def _write_timestamp_fractional_seconds(buf, value):
    sign, digits, exponent = value.as_tuple()
    coefficient = int(value.scaleb(-exponent).to_integral_value())
    if coefficient == 0 and exponent >= 0:
        length = 0
    else:
        length = _write_decimal_value(buf, exponent, coefficient, sign)
    return length


def _serialize_decimal(ion_event):
    buf = bytearray()
    value = ion_event.value
    validate_scalar_value(value, Decimal)
    sign, digits, exponent = value.as_tuple()
    with localcontext() as context:
        # Adjusting precision for taking into account arbitrarily large/small
        # numbers
        context.prec = len(digits)
        coefficient = int(value.scaleb(-exponent).to_integral_value())
    if not sign and not exponent and not coefficient:
        # The value is 0d0; other forms of zero will fall through.
        buf.append(_Zeros.DECIMAL)
    else:
        value_buf = bytearray()
        length = _write_decimal_value(value_buf, exponent, coefficient, sign)
        _write_length(buf, length, _TypeIds.DECIMAL)
        buf.extend(value_buf)
    return buf


def _serialize_string(ion_event):
    buf = bytearray()
    value = ion_event.value
    validate_scalar_value(value, str)
    if not value:
        buf.append(_Zeros.STRING)
    else:
        value_buf = value.encode('utf-8')
        _write_length(buf, len(value_buf), _TypeIds.STRING)
        buf.extend(value_buf)
    return buf


def _serialize_symbol(ion_event):
    buf = bytearray()
    token = ion_event.value
    validate_scalar_value(token, SymbolToken)
    sid = token.sid
    if sid == 0:
        buf.append(_Zeros.SYMBOL)
    else:
        _write_int_value(buf, _TypeIds.SYMBOL, sid)
    return buf


def _serialize_lob_value(event, tid):
    buf = bytearray()
    value = event.value
    _write_length(buf, len(value), tid)
    buf.extend(value)
    return buf


_serialize_blob = partial(_serialize_lob_value, tid=_TypeIds.BLOB)
_serialize_clob = partial(_serialize_lob_value, tid=_TypeIds.CLOB)


def _serialize_timestamp(ion_event):
    buf = bytearray()
    dt = ion_event.value
    precision = getattr(dt, TIMESTAMP_PRECISION_FIELD, TimestampPrecision.SECOND)
    if precision is None:  # TODO should this defaulting be pushed into Timestamp itself?
        precision = TimestampPrecision.SECOND
    validate_scalar_value(dt, datetime)
    value_buf = bytearray()
    if dt.tzinfo is None:
        value_buf.append(_VARINT_NEG_ZERO)  # This signifies an unknown local offset.
        length = 1
    else:
        # Normalize to UTC and write the offset field.
        offset = dt.utcoffset()
        dt -= offset
        length = _write_varint(value_buf, int(total_seconds(offset) // 60))
    length += _write_varuint(value_buf, dt.year)
    if precision.includes_month:
        length += _write_varuint(value_buf, dt.month)
    if precision.includes_day:
        length += _write_varuint(value_buf, dt.day)
    if precision.includes_minute:
        length += _write_varuint(value_buf, dt.hour)
        length += _write_varuint(value_buf, dt.minute)
    if precision.includes_second:
        length += _write_varuint(value_buf, dt.second)
        if isinstance(ion_event.value, Timestamp):
            fractional_seconds = getattr(ion_event.value, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, None)
            if fractional_seconds is not None:
                length += _write_timestamp_fractional_seconds(value_buf, fractional_seconds)
        else:
            # This must be a normal datetime, which always has a range-validated microsecond value.
            length += _write_decimal_value(value_buf, -MICROSECOND_PRECISION, dt.microsecond)

    _write_length(buf, length, _TypeIds.TIMESTAMP)
    buf.extend(value_buf)
    return buf


_SERIALIZE_SCALAR_JUMP_TABLE = {
    IonType.NULL: illegal_state_null,
    IonType.BOOL: _serialize_bool,
    IonType.INT: _serialize_int,
    IonType.FLOAT: _serialize_float,
    IonType.DECIMAL: _serialize_decimal,
    IonType.TIMESTAMP: _serialize_timestamp,
    IonType.SYMBOL: _serialize_symbol,
    IonType.STRING: _serialize_string,
    IonType.CLOB: _serialize_clob,
    IonType.BLOB: _serialize_blob,
}


_serialize_scalar = partial(
    serialize_scalar, jump_table=_SERIALIZE_SCALAR_JUMP_TABLE, null_table=_NULLS
)


def _serialize_annotation_wrapper(output_buf, annotations):
    value_length = output_buf.current_container_length
    annot_length_buf = bytearray()
    annot_length = 0
    for annotation in annotations:
        annot_length += _write_varuint(annot_length_buf, annotation.sid)
    header = bytearray()
    length_buf = bytearray()
    length = _write_varuint(length_buf, annot_length) + annot_length + value_length
    _write_length(header, length, _TypeIds.ANNOTATION_WRAPPER)
    header.extend(length_buf)
    header.extend(annot_length_buf)
    output_buf.end_container(header)


def _serialize_container(output_buf, ion_event):
    ion_type = ion_event.ion_type
    length = output_buf.current_container_length
    header = bytearray()
    if ion_type is IonType.STRUCT:
        if length == 0:
            header.append(_Zeros.STRUCT)
        else:
            # TODO Support sorted field name symbols, per the spec.
            header.append(_TypeIds.STRUCT | _LENGTH_FIELD_INDICATOR)
            _write_varuint(header, length)
    else:
        tid = _TypeIds.LIST
        if ion_type is IonType.SEXP:
            tid = _TypeIds.SEXP
        _write_length(header, length, tid)
    output_buf.end_container(header)


_WRITER_EVENT_NEEDS_INPUT_EMPTY = DataEvent(WriteEventType.NEEDS_INPUT, b'')


@coroutine
def _raw_writer_coroutine(writer_buffer, depth=0, container_event=None,
                          whence=None, pending_annotations=None):

    def fail():
        raise TypeError('Invalid event: %s at depth %d' % (ion_event, depth))

    write_result = None
    while True:
        ion_event, self = (yield write_result)
        delegate = self
        curr_annotations = ion_event.annotations
        writer_event = _WRITER_EVENT_NEEDS_INPUT_EMPTY
        if depth > 0 and container_event.ion_type is IonType.STRUCT \
                and ion_event.event_type.begins_value:
            # A field name symbol ID is required at this position.
            sid_buffer = bytearray()
            _write_varuint(sid_buffer, ion_event.field_name.sid)  # Write the field name's symbol ID.
            writer_buffer.add_scalar_value(sid_buffer)
        if ion_event.event_type.begins_value and curr_annotations:
            writer_buffer.start_container()
        if ion_event.event_type is IonEventType.SCALAR:
            scalar_buffer = _serialize_scalar(ion_event)
            writer_buffer.add_scalar_value(scalar_buffer)
            if curr_annotations:
                _serialize_annotation_wrapper(writer_buffer, curr_annotations)
        elif ion_event.event_type is IonEventType.STREAM_END:
            if depth != 0:
                fail()
            for partial_value in writer_buffer.drain():
                yield partial_transition(partial_value, self)
            writer_event = NOOP_WRITER_EVENT
        elif ion_event.event_type is IonEventType.CONTAINER_START:
            if not ion_event.ion_type.is_container:
                raise TypeError('Expected container type')
            writer_buffer.start_container()
            delegate = _raw_writer_coroutine(writer_buffer, depth + 1,
                                             ion_event, self, curr_annotations)
        elif ion_event.event_type is IonEventType.CONTAINER_END:
            if depth < 1:
                fail()
            _serialize_container(writer_buffer, container_event)
            if pending_annotations:
                _serialize_annotation_wrapper(writer_buffer, pending_annotations)
            pending_annotations = None
            delegate = whence
        else:
            fail()
        write_result = Transition(writer_event, delegate)


def _raw_binary_writer(writer_buffer):
    """Returns a raw binary writer co-routine.

    Yields:
        DataEvent: serialization events to write out

        Receives :class:`amazon.ion.core.IonEvent`.
    """
    return writer_trampoline(_raw_writer_coroutine(writer_buffer))



from typing import NamedTuple, List


class SliceableBuffer:
    """
    A theoretically infinite immutable buffer from which readers may read bytes
    or slices by position.

    Reads return the data and a new buffer starting at the end of the read.
    As the reader advances past chunks (either through reads or skips), whole
    chunks are dropped from the buffer.

    Built with the assumption that chunks will be reasonably large and that
    relatively few (single digit) chunks will be buffered at once.
    """

    @staticmethod
    def empty():
        """
        Create a new buffer with no data.
        """
        return SliceableBuffer([])

    def __init__(self, chunks, offset=0, size=0):
        """
        *Class internal usage only.*

        Users should use `empty()` to get a new buffer.
        """
        self._chunks: List[_ChunkPair] = chunks
        # the offset adds complexity but enables keeping and dropping of whole
        # chunks which is more efficient than slicing and copying the chunks
        # on each read or skip.
        self._offset = offset
        self.size = size

    def extend(self, chunk):
        """
        Return a new buffer with the chunk appended.
        """
        if not chunk:
            raise ValueError("Chunk must be not None and non-empty!")

        mem_chunk = memoryview(chunk)
        pair = _ChunkPair(mem_chunk, len(mem_chunk))
        return SliceableBuffer(
            self._chunks + [pair],
            self._offset,
            self.size + pair.length)

    def read_byte(self):
        """
        Read the next byte from the buffer, return (byte, new buffer).

        Raise IncompleteReadError if the buffer is empty.
        """
        size = self.size
        chunks = self._chunks
        offset = self._offset

        try:
            # assume that we have data, and that chunks are non-empty
            (chunk, length) = chunks[0]
        except IndexError:
            raise IncompleteReadError("Buffer is empty!")

        if length == offset + 1:
            return chunk[offset], SliceableBuffer(chunks[1:], 0, size - 1)
        else:
            return chunk[offset], SliceableBuffer(chunks, offset + 1, size - 1),

    def read_slice(self, n):
        """
        Read a slice of the buffer, return (slice, new buffer).

        Raise IncompleteReadError if the slice cannot be fully read.

        Chunks which are no longer readable are dropped from the new buffer.
        Bytes are only copied if the read requires bridging chunks.
        """
        size = self.size
        chunks = self._chunks
        offset = self._offset
        if n < 1:
            raise ValueError("n must be >= 1")

        endpos = offset + n

        if size < n:
            raise IncompleteReadError(f'Buffer has size {size}, but {n} bytes were requested!')

        # short-circuit when we can serve full read from first chunk
        # optimizes for common case and simplifies accumulation loop
        (chunk, length) = chunks[0]
        if endpos < length:
            return chunk[offset:endpos], SliceableBuffer(chunks, offset + n, size - n)
        elif endpos == length:
            return chunk[offset:], SliceableBuffer(chunks[1:], 0, size - n)

        slices = [_ChunkPair(chunk[offset:], length - offset)]

        # remaining and i are used to init the new buffer after the loop
        remaining = endpos - length
        i = 1
        for (i, pair) in enumerate(chunks[1:], start=1):
            (chunk, length) = pair
            if remaining < length:
                slices.append(_ChunkPair(chunk[:remaining], remaining))
                break

            slices.append(pair)
            remaining -= length
            if remaining == 0:
                # move i forward to drop the chunk in the new buffer
                i += 1
                break

        combined = bytearray(n)
        cursor = 0
        for (chunk, length) in slices:
            combined[cursor:cursor + length] = chunk
            cursor += length

        return memoryview(combined), SliceableBuffer(chunks[i:], remaining, size - n)

    def skip(self, n):
        """
        Skip max(n, size) bytes, return (skipped, new buffer).

        Chunks which are no longer readable are dropped from the new buffer.

        Unlike the read methods, skip allows partial skipping, which is more
        memory efficient when skipping large tokens that span chunks.
        """
        size = self.size
        chunks = self._chunks
        offset = self._offset
        endpos = offset + n

        if size <= n:
            return size, SliceableBuffer([])

        remaining = endpos
        i = 0
        for (i, (_, length)) in enumerate(chunks):
            if remaining < length:
                break

            remaining -= length
            if remaining == 0:
                # move i forward to drop the chunk in the new buffer
                i += 1
                break

        return n, SliceableBuffer(chunks[i:], remaining, size - n)

    def __len__(self):
        """
        Length of data in bytes remaining in buffer.
        """
        return self.size


class IncompleteReadError(IndexError):
    pass


class _ChunkPair(NamedTuple):
    chunk: memoryview
    length: int



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Implementations of Ion Text writers."""
import base64
import math
import re
from functools import partial

from datetime import datetime
from decimal import Decimal
from io import BytesIO

from amazon.ion.symbols import SymbolToken
from . import symbols

from .util import coroutine, unicode_iter
from .core import DataEvent, Transition, IonEventType, IonType, TIMESTAMP_PRECISION_FIELD, TimestampPrecision, \
    _ZERO_DELTA, TIMESTAMP_FRACTION_PRECISION_FIELD, MICROSECOND_PRECISION, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, \
    Timestamp, DECIMAL_ZERO
from .writer import partial_transition, writer_trampoline, serialize_scalar, validate_scalar_value, \
    illegal_state_null, NOOP_WRITER_EVENT
from .writer import WriteEventType

_IVM_WRITER_EVENT = DataEvent(WriteEventType.COMPLETE, symbols.TEXT_ION_1_0.encode())

_NULL_TYPE_NAMES = [
    b'null',
    b'null.bool',
    b'null.int',
    b'null.float',
    b'null.decimal',
    b'null.timestamp',
    b'null.symbol',
    b'null.string',
    b'null.clob',
    b'null.blob',
    b'null.list',
    b'null.sexp',
    b'null.struct',
]


def _serialize_bool(ion_event):
    if ion_event.value:
        return b'true'
    else:
        return b'false'


def _serialize_scalar_from_string_representation_factory(type_name, types, str_func=str):
    """Builds functions that leverage Python ``str()`` or similar functionality.

    Args:
        type_name (str): The name of the Ion type.
        types (Union[Sequence[type],type]): The Python types to validate for.
        str_func (Optional[Callable]): The function to convert the value with, defaults to ``str``.

    Returns:
        function: The function for serializing scalars of a given type to Ion text bytes.
    """
    def serialize(ion_event):
        value = ion_event.value
        validate_scalar_value(value, types)

        # This assumes an encoding of UTF-8, which is the only one supported at time
        # of writing (Dec 2022).
        return str_func(value).encode("utf-8")
    serialize.__name__ = '_serialize_' + type_name
    return serialize


_serialize_int = _serialize_scalar_from_string_representation_factory(
    'int', int
)


_EXPONENT_PAT = re.compile(r'[eE]')


# TODO Make this cleaner.
def _float_str(val):
    if math.isnan(val):
        return 'nan'
    if math.isinf(val):
        if val > 0:
            return '+inf'
        else:
            return '-inf'
    text = repr(val)
    if _EXPONENT_PAT.search(text) is None:
        text += 'e0'
    return text

_serialize_float = _serialize_scalar_from_string_representation_factory(
    'float', float,
    str_func=_float_str
)


# TODO Make this cleaner.
def _decimal_str(val):
    text = str(val)
    new_text = _EXPONENT_PAT.sub('d', text)
    if text == new_text and text.find('.') == -1:
        new_text += 'd0'
    return new_text


_serialize_decimal = _serialize_scalar_from_string_representation_factory(
    'decimal', Decimal,
    str_func=_decimal_str
)


def _bytes_utc_offset(dt):
    offset = dt.utcoffset()
    if offset is None:
        return '-00:00'
    elif offset == _ZERO_DELTA:
        return 'Z'
    offset_str = dt.strftime('%z')
    offset_str = offset_str[:3] + ':' + offset_str[3:]
    return offset_str


def _bytes_datetime(dt):
    original_dt = dt
    precision = getattr(original_dt, TIMESTAMP_PRECISION_FIELD, TimestampPrecision.SECOND)
    if dt.year < 1900:
        # In some Python interpreter versions, strftime inexplicably does not support pre-1900 years.
        # This unfortunate ugliness compensates for that.
        year = str(dt.year)
        year = ('0' * (4 - len(year))) + year
        dt = dt.replace(year=2008)  # Note: this fake year must be a leap year.
    else:
        year = dt.strftime('%Y')
    tz_string = year

    if precision.includes_month:
        tz_string += dt.strftime('-%m')
    else:
        return tz_string + 'T'

    if precision.includes_day:
        tz_string += dt.strftime('-%dT')
    else:
        return tz_string + 'T'

    if precision.includes_minute:
        tz_string += dt.strftime('%H:%M')
    else:
        return tz_string

    if precision.includes_second:
        tz_string += dt.strftime(':%S')
    else:
        return tz_string + _bytes_utc_offset(dt)

    if isinstance(original_dt, Timestamp):
        fractional_seconds = getattr(original_dt, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, None)
        if fractional_seconds is not None:
            _, digits, exponent = fractional_seconds.as_tuple()
            if not (fractional_seconds == DECIMAL_ZERO and exponent >= 0):
                leading_zeroes = -exponent - len(digits)
                tz_string += '.'
                if leading_zeroes > 0:
                    tz_string += '0' * leading_zeroes
                tz_string += ''.join(str(x) for x in digits)
    else:
        # This must be a normal datetime, which always has a range-validated microsecond value.
        tz_string += '.' + dt.strftime('%f')
    return tz_string + _bytes_utc_offset(dt)


_serialize_timestamp = _serialize_scalar_from_string_representation_factory(
    'timestamp',
    datetime,
    str_func=_bytes_datetime
)


_PRINTABLE_ASCII_START = 0x20
_PRINTABLE_ASCII_END = 0x7E


def _is_printable_ascii(code_point):
    return code_point >= _PRINTABLE_ASCII_START and code_point <= _PRINTABLE_ASCII_END


_SERIALIZE_COMMON_ESCAPE_MAP = {
    b'\n'[0]: br'\n',
    b'\r'[0]: br'\r',
    b'\t'[0]: br'\t',
}
_2B_ESCAPE_MAX = 0xFF
_4B_ESCAPE_MAX = 0xFFFF


def _escape(code_point):
    escape = _SERIALIZE_COMMON_ESCAPE_MAP.get(code_point, None)
    if escape is not None:
        return escape
    if code_point <= _2B_ESCAPE_MAX:
        return (u'\\x%02x' % code_point).encode()
    if code_point <= _4B_ESCAPE_MAX:
        return (u'\\u%04x' % code_point).encode()
    return (u'\\U%08x' % code_point).encode()


def _bytes_text(code_point_iter, quote, prefix=b'', suffix=b''):
    quote_code_point = None if len(quote) == 0 else quote[0]
    buf = BytesIO()
    buf.write(prefix)
    buf.write(quote)
    for code_point in code_point_iter:
        if code_point == quote_code_point:
            buf.write(b'\\' + quote)
        elif code_point == b'\\'[0]:
            buf.write(b'\\\\')
        elif _is_printable_ascii(code_point):
            buf.write(bytes((code_point,)))
        else:
            buf.write(_escape(code_point))
    buf.write(quote)
    buf.write(suffix)
    return buf.getvalue()


_SINGLE_QUOTE = b"'"
_DOUBLE_QUOTE = b'"'
# all typed nulls (such as null.int) and the +inf, and -inf keywords are covered by this regex
_UNQUOTED_SYMBOL_REGEX = re.compile(r'\A[a-zA-Z$_][a-zA-Z0-9$_]*\Z')
_ADDITIONAL_SYMBOLS_REQUIRING_QUOTES = set(['nan', 'null', 'false', 'true'])

def _symbol_needs_quotes(text):
    return text in _ADDITIONAL_SYMBOLS_REQUIRING_QUOTES or _UNQUOTED_SYMBOL_REGEX.search(text) is None

def _serialize_symbol_value(value, suffix=b''):
    # TODO Support not quoting operators in s-expressions: https://amazon-ion.github.io/ion-docs/docs/symbols.html
    try:
        text = value.text
        if text is None:
            return (u'$%d' % value.sid).encode() + suffix
    except AttributeError:
        text = value
    validate_scalar_value(text, (str, type(SymbolToken)))
    quote = _SINGLE_QUOTE if _symbol_needs_quotes(text) else b''
    return _bytes_text(unicode_iter(text), quote, suffix=suffix)


def _serialize_symbol(ion_event):
    return _serialize_symbol_value(ion_event.value)


def _serialize_string(ion_event):
    # TODO Support multi-line strings.
    value = ion_event.value
    validate_scalar_value(value, str)
    return _bytes_text(unicode_iter(value), _DOUBLE_QUOTE)


_LOB_START = b'{{'
_LOB_END = b'}}'


def _serialize_clob(ion_event):
    value = ion_event.value
    return _bytes_text(iter(value), _DOUBLE_QUOTE, prefix=_LOB_START, suffix=_LOB_END)


def _serialize_blob(ion_event):
    value = ion_event.value
    return _LOB_START + base64.b64encode(value) + _LOB_END


_SERIALIZE_SCALAR_JUMP_TABLE = {
    IonType.NULL: illegal_state_null,
    IonType.BOOL: _serialize_bool,
    IonType.INT: _serialize_int,
    IonType.FLOAT: _serialize_float,
    IonType.DECIMAL: _serialize_decimal,
    IonType.TIMESTAMP: _serialize_timestamp,
    IonType.SYMBOL: _serialize_symbol,
    IonType.STRING: _serialize_string,
    IonType.CLOB: _serialize_clob,
    IonType.BLOB: _serialize_blob,
}


_serialize_scalar = partial(serialize_scalar, jump_table=_SERIALIZE_SCALAR_JUMP_TABLE, null_table=_NULL_TYPE_NAMES)


_FIELD_NAME_DELIMITER = b':'
_ANNOTATION_DELIMITER = b'::'


def _serialize_field_name(ion_event):
    return _serialize_symbol_value(ion_event.field_name, suffix=_FIELD_NAME_DELIMITER)


def _serialize_annotation_value(annotation):
    return _serialize_symbol_value(annotation, suffix=_ANNOTATION_DELIMITER)


def _serialize_container_factory(suffix, container_map):
    """Returns a function that serializes container start/end.

    Args:
        suffix (str): The suffix to name the function with.
        container_map (Dictionary[core.IonType, bytes]): The

    Returns:
        function: The closure for serialization.
    """
    def serialize(ion_event):
        if not ion_event.ion_type.is_container:
            raise TypeError('Expected container type')
        return container_map[ion_event.ion_type]
    serialize.__name__ = '_serialize_container_' + suffix
    return serialize

_CONTAINER_START_MAP = {
    IonType.STRUCT: b'{',
    IonType.LIST: b'[',
    IonType.SEXP: b'(',
}
_CONTAINER_END_MAP = {
    IonType.STRUCT: b'}',
    IonType.LIST: b']',
    IonType.SEXP: b')',
}
_CONTAINER_DELIMITER_MAP_NORMAL = {
    IonType.STRUCT: b',',
    IonType.LIST: b',',
    IonType.SEXP: b' ',
}
_CONTAINER_DELIMITER_MAP_PRETTY = {
    IonType.STRUCT: b',',
    IonType.LIST: b',',
    IonType.SEXP: b'', # we use newlines when pretty printing
}

_serialize_container_start = _serialize_container_factory('start', _CONTAINER_START_MAP)
_serialize_container_end = _serialize_container_factory('end', _CONTAINER_END_MAP)
_serialize_container_delimiter_normal = _serialize_container_factory('delimiter', _CONTAINER_DELIMITER_MAP_NORMAL)
_serialize_container_delimiter_pretty = _serialize_container_factory('delimiter', _CONTAINER_DELIMITER_MAP_PRETTY)


@coroutine
def _raw_writer_coroutine(depth=0, container_event=None, whence=None, indent=None):
    pretty = indent is not None
    serialize_container_delimiter = \
            _serialize_container_delimiter_pretty if pretty else _serialize_container_delimiter_normal
    has_written_values = False
    transition = None
    while True:
        ion_event, self = (yield transition)
        delegate = self

        if has_written_values and not ion_event.event_type.ends_container:
            # TODO This will always emit a delimiter for containers--should make it not do that.
            # Write the delimiter for the next value.
            if depth == 0:
                # if we are pretty printing, we'll insert a newline between top-level containers
                delimiter = b'' if pretty else b' '
            else:
                delimiter = serialize_container_delimiter(container_event)
            if len(delimiter) > 0:
                yield partial_transition(delimiter, self)

        if pretty and (has_written_values or container_event is not None) and not ion_event.event_type is IonEventType.STREAM_END:
            yield partial_transition(b'\n', self)
            indent_depth = depth - (1 if ion_event.event_type is IonEventType.CONTAINER_END else 0)
            if indent_depth > 0:
                yield partial_transition(indent * indent_depth, self)

        if depth > 0 \
                and container_event.ion_type is IonType.STRUCT \
                and ion_event.event_type.begins_value:
            # Write the field name.
            yield partial_transition(_serialize_field_name(ion_event), self)
            if pretty:
                # separate the field name and the field value
                yield partial_transition(b' ', self)

        if ion_event.event_type.begins_value:
            # Write the annotations.
            for annotation in ion_event.annotations:
                yield partial_transition(_serialize_annotation_value(annotation), self)

        if ion_event.event_type is IonEventType.CONTAINER_START:
            writer_event = DataEvent(WriteEventType.NEEDS_INPUT, _serialize_container_start(ion_event))
            delegate = _raw_writer_coroutine(depth + 1, ion_event, self, indent=indent)
        elif depth == 0:
            # Serialize at the top-level.
            if ion_event.event_type is IonEventType.STREAM_END:
                writer_event = NOOP_WRITER_EVENT
            elif ion_event.event_type is IonEventType.VERSION_MARKER:
                writer_event = _IVM_WRITER_EVENT
            elif ion_event.event_type is IonEventType.SCALAR:
                writer_event = DataEvent(WriteEventType.COMPLETE, _serialize_scalar(ion_event))
            else:
                raise TypeError('Invalid event: %s' % ion_event)
        else:
            # Serialize within a container.
            if ion_event.event_type is IonEventType.SCALAR:
                writer_event = DataEvent(WriteEventType.NEEDS_INPUT, _serialize_scalar(ion_event))
            elif ion_event.event_type is IonEventType.CONTAINER_END:
                write_type = WriteEventType.COMPLETE if depth == 1 else WriteEventType.NEEDS_INPUT
                writer_event = DataEvent(write_type, _serialize_container_end(container_event))
                delegate = whence
            else:
                raise TypeError('Invalid event: %s' % ion_event)

        has_written_values = True
        transition = Transition(writer_event, delegate)


# TODO Add options for text formatting.
def raw_writer(indent=None):
    """Returns a raw text writer co-routine.

    Yields:
        DataEvent: serialization events to write out

        Receives :class:`amazon.ion.core.IonEvent` or ``None`` when the co-routine yields
        ``HAS_PENDING`` :class:`WriteEventType` events.
    """

    is_whitespace_str = isinstance(indent, str) and re.search(r'\A\s*\Z', indent, re.M) is not None
    if not (indent is None or is_whitespace_str):
        raise ValueError('The indent parameter must either be None or a string containing only whitespace')

    # This assumes an encoding of UTF-8, which is the only one supported at time
    # of writing (Dec 2022).
    indent_bytes = indent.encode("UTF-8") if isinstance(indent, str) else indent

    return writer_trampoline(_raw_writer_coroutine(indent=indent_bytes))

# TODO Determine if we need to do anything special for non-raw writer.  Validation?
text_writer = raw_writer



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.
from enum import IntEnum
from typing import NamedTuple, Optional, Any, Union, Sequence, Coroutine

from amazon.ion.symbols import SymbolToken

# in Python 3.10, abstract collections have moved into their own module
# for compatibility with 3.10+, first try imports from the new location
# if that fails, try from the pre-3.10 location
try:
    from collections.abc import MutableMapping, MutableSequence
    from collections import OrderedDict
except:
    from collections import MutableMapping, MutableSequence, OrderedDict

from datetime import datetime, timedelta, tzinfo
from decimal import Decimal, Context, Inexact
from math import isnan


class IonType(IntEnum):
    """Enumeration of the Ion data types."""
    NULL = 0
    BOOL = 1
    INT = 2
    FLOAT = 3
    DECIMAL = 4
    TIMESTAMP = 5
    SYMBOL = 6
    STRING = 7
    CLOB = 8
    BLOB = 9
    LIST = 10
    SEXP = 11
    STRUCT = 12

    @property
    def is_numeric(self):
        return IonType.INT <= self <= IonType.TIMESTAMP

    @property
    def is_text(self):
        """Returns whether the type is a Unicode textual type."""
        return self is IonType.SYMBOL or self is IonType.STRING

    @property
    def is_lob(self):
        """Returns whether the type is a LOB."""
        return self is IonType.CLOB or self is IonType.BLOB

    @property
    def is_container(self):
        """Returns whether the type is a container."""
        return self >= IonType.LIST


# TODO At some point we can add SCALAR_START/SCALAR_END for streaming large values.
class IonEventType(IntEnum):
    """Enumeration of Ion parser or serializer events.

    These types do not correspond directly to the Ion type system, but they are related.
    In particular, ``null.*`` will surface as a ``SCALAR`` even though they are containers.

    Attributes:
        INCOMPLETE: Indicates that parsing cannot be completed due to lack of input.
        STREAM_END: Indicates that the logical stream has terminated.
        VERSION_MARKER: Indicates that the **Ion Version Marker** has been encountered.
        SCALAR: Indicates an *atomic* value has been encountered.
        CONTAINER_START: Indicates that the start of a container has been reached.
        CONTAINER_END: Indicates that the end of a container has been reached.
    """
    INCOMPLETE = -2
    STREAM_END = -1
    VERSION_MARKER = 0
    SCALAR = 1
    CONTAINER_START = 2
    CONTAINER_END = 3

    @property
    def begins_value(self):
        """Indicates if the event type is a start of a value."""
        return self is IonEventType.SCALAR or self is IonEventType.CONTAINER_START

    @property
    def ends_container(self):
        """Indicates if the event type terminates a container or stream."""
        return self is IonEventType.STREAM_END or self is IonEventType.CONTAINER_END

    @property
    def is_stream_signal(self):
        """Indicates that the event type corresponds to a stream signal."""
        return self < 0


SymbolTokenOrStr = Union[SymbolToken, str]


class IonEvent(NamedTuple):
    """A parse or serialization event.

    Args:
        event_type (IonEventType): The type of event.
        ion_type (Optional(amazon.ion.core.IonType)): The Ion data model type
            associated with the event.
        value (Optional[any]): The data value associated with the event.
        field_name (Optional[SymbolTokenOrStr]): The field name
            associated with the event.
        annotations (Sequence[SymbolTokenOrStr]): The annotations
            associated with the event.
        depth (Optional[int]): The tree depth of the event if applicable.
    """
    event_type: IonEventType
    ion_type: Optional[IonType] = None
    value: Optional[Any] = None
    field_name: Optional[SymbolTokenOrStr] = None
    annotations: Sequence[SymbolTokenOrStr] = ()
    depth: Optional[int] = None

    def __eq__(self, other):
        if not isinstance(other, IonEvent):
            return False

        if isinstance(self.value, float):
            if not isinstance(other.value, float):
                return False

            # Need to deal with NaN appropriately.
            if self.value != other.value and not (isnan(self.value) and isnan(other.value)):
                return False
        else:
            if self.value != other.value:
                return False

            # Timestamp precision has additional requirements.
            if isinstance(self.value, Timestamp) or isinstance(other.value, Timestamp):
                # Special case for timestamps to capture equivalence over precision.
                self_precision = getattr(self.value, TIMESTAMP_PRECISION_FIELD, None)
                other_precision = getattr(other.value, TIMESTAMP_PRECISION_FIELD, None)
                if self_precision != other_precision \
                        and not ((self_precision is None and other_precision is TimestampPrecision.SECOND) or
                                 (self_precision is TimestampPrecision.SECOND and other_precision is None)):
                    # The absence of precision indicates a naive datetime, which always has SECOND precision.
                    return False

            if isinstance(self.value, Timestamp) and isinstance(other.value, Timestamp):
                self_fractional_seconds = getattr(self.value, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, None)
                other_fractional_seconds = getattr(other.value, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, None)
                if self_fractional_seconds != other_fractional_seconds:
                    return False

            if isinstance(self.value, datetime):
                if self.value.utcoffset() != other.value.utcoffset():
                    return False

        return (self.event_type == other.event_type
            and self.ion_type == other.ion_type
            and self.field_name == other.field_name
            and self.annotations == other.annotations
            and self.depth == other.depth
        )

    def derive_field_name(self, field_name):
        """Derives a new event from this one setting the ``field_name`` attribute.

        Args:
            field_name (Union[amazon.ion.symbols.SymbolToken, unicode]): The field name to set.
        Returns:
            IonEvent: The newly generated event.
        """
        cls = type(self)
        # We use ordinals to avoid thunk materialization.
        return cls(
            self[0],
            self[1],
            self[2],
            field_name,
            self[4],
            self[5]
        )

    def derive_annotations(self, annotations):
        """Derives a new event from this one setting the ``annotations`` attribute.

        Args:
            annotations: (Sequence[Union[amazon.ion.symbols.SymbolToken, unicode]]):
                The annotations associated with the derived event.

        Returns:
            IonEvent: The newly generated event.
        """
        cls = type(self)
        # We use ordinals to avoid thunk materialization.
        return cls(
            self[0],
            self[1],
            self[2],
            self[3],
            annotations,
            self[5]
        )

    def derive_value(self, value):
        """Derives a new event from this one setting the ``value`` attribute.

        Args:
            value: (any):
                The value associated with the derived event.

        Returns:
            IonEvent: The newly generated non-thunk event.
        """
        return IonEvent(
            self.event_type,
            self.ion_type,
            value,
            self.field_name,
            self.annotations,
            self.depth
        )

    def derive_depth(self, depth):
        """Derives a new event from this one setting the ``depth`` attribute.

        Args:
            depth: (int):
                The annotations associated with the derived event.

        Returns:
            IonEvent: The newly generated event.
        """
        cls = type(self)
        # We use ordinals to avoid thunk materialization.
        return cls(
            self[0],
            self[1],
            self[2],
            self[3],
            self[4],
            depth
        )


class MemoizingThunk(object):
    """A :class:`callable` that invokes a ``delegate`` and caches and returns the result."""
    def __init__(self, delegate):
        self.delegate = delegate

    def __call__(self):
        if hasattr(self, 'value'):
            return self.value
        self.value = self.delegate()
        return self.value

    def __str__(self):
        return str(self())

    def __repr__(self):
        return repr(self())


class IonThunkEvent(IonEvent):
    """
    A lazy `IonEvent` whose ``value`` field is a thunk.

    The `value` will be materialized on first access and cached.

    Accessing the value by its slot: ``event[2]`` will avoid materialization.
    """

    @property
    def value(self):
        if hasattr(self, 'cached_value'):
            return self.cached_value
        self.cached_value = self[2]()
        return self.cached_value

# Singletons for structural events
ION_STREAM_END_EVENT = IonEvent(IonEventType.STREAM_END)
ION_STREAM_INCOMPLETE_EVENT = IonEvent(IonEventType.INCOMPLETE)
ION_VERSION_MARKER_EVENT = IonEvent(
    IonEventType.VERSION_MARKER, ion_type=None, value=(1, 0), depth=0
)


class DataEvent(NamedTuple):
    """Event generated as a result of the writer or as input into the reader.

    Args:
        type (IntEnum): The type of event.
        data (bytes):  The serialized data returned.  If no data is to be serialized,
            this should be the empty byte string.
    """
    type: IntEnum
    data: bytes


class Transition(NamedTuple):
    """A pair of event and co-routine delegate.

    This is generally used as a result of a state-machine.

    Args:
        event (Union[DataEvent, IonEvent, None]): The event associated with the transition.
        delegate (Coroutine): The co-routine delegate which can be the same routine from
            whence this transition came.
    """
    event: Union[DataEvent, IonEvent, None]
    delegate: Coroutine


_MIN_OFFSET = timedelta(hours=-24)
_MAX_OFFSET = timedelta(hours=24)
_ZERO_DELTA = timedelta()


class OffsetTZInfo(tzinfo):
    """A trivial UTC offset :class:`tzinfo`."""
    def __init__(self, delta=_ZERO_DELTA):
        if delta <= _MIN_OFFSET or delta >= _MAX_OFFSET:
            raise ValueError('Invalid UTC offset: %s' % delta)
        self.delta = delta

    def dst(self, date_time):
        return timedelta()

    def tzname(self, date_time):
        return None

    def utcoffset(self, date_time):
        return self.delta

    def __repr__(self):
        sign = '+'
        delta = self.delta
        if delta < _ZERO_DELTA:
            sign = '-'
            delta = _ZERO_DELTA - delta
        return 'OffsetTZInfo(%s%s)' % (sign, delta)


class TimestampPrecision(IntEnum):
    """The different levels of precision supported in an Ion timestamp."""
    YEAR = 0
    MONTH = 1
    DAY = 2
    MINUTE = 3
    SECOND = 4

    @property
    def includes_month(self):
        """Precision has at least the ``month`` field."""
        return self >= TimestampPrecision.MONTH

    @property
    def includes_day(self):
        """Precision has at least the ``day`` field."""
        return self >= TimestampPrecision.DAY

    @property
    def includes_minute(self):
        """Precision has at least the ``minute`` field."""
        return self >= TimestampPrecision.MINUTE

    @property
    def includes_second(self):
        """Precision has at least the ``second`` field."""
        return self >= TimestampPrecision.SECOND


TIMESTAMP_PRECISION_FIELD = 'precision'
TIMESTAMP_FRACTION_PRECISION_FIELD = 'fractional_precision'
TIMESTAMP_FRACTIONAL_SECONDS_FIELD = 'fractional_seconds'
TIMESTAMP_MICROSECOND_FIELD = 'microsecond'
MICROSECOND_PRECISION = 6
BASE_TEN_MICROSECOND_PRECISION_EXPONENTIATION = 10 ** MICROSECOND_PRECISION
DECIMAL_ZERO = Decimal(0)
PRECISION_LIMIT_LOOKUP = (
    DECIMAL_ZERO,
    Decimal('0.1'),
    Decimal('0.01'),
    Decimal('0.001'),
    Decimal('0.0001'),
    Decimal('0.00001'),
    Decimal('0.000001')
)
DATETIME_CONSTRUCTOR_MICROSECOND_ARGUMENT_INDEX = 6


class Timestamp(datetime):
    """Sub-class of :class:`datetime` that supports a precision field; a ``fractional_precision``
        field that specifies the  precision of the``microseconds`` field in :class:`datetime`;
        and a ``fractional_seconds`` field that is a :class:`Decimal` specifying the fractional
        seconds precisely.

        Notes:
            * The ``precision`` field is passed as a keyword argument of the same name.

            * The ``fractional_precision`` field is passed as a keyword argument of the same name.
              This field only relates to the ``microseconds`` field and can be thought of
              as the number of decimal digits that are significant.  This is an integer that
              that is in the closed interval ``[0, 6]``.  If ``0``, ``microseconds`` must be
              ``0`` indicating no precision below seconds.  This argument is optional and only valid
              when ``microseconds`` is not ``None``.  If the ``microseconds`` specified has more
              precision than this field indicates, then that is an error.

            * The ``fractional_seconds`` field is passed as a keyword argument of the same name.
              It must be a :class:`Decimal` in the left-closed, right-opened interval of ``[0, 1)``.
              If specified as an argument, ``microseconds`` must be ``None`` **and** ``fractional_precision``
              must not be specified (but can be ``None``).  In addition, if ``microseconds`` is specified
              this argument must not be specified (but can be ``None``). If the specified value has
              ``coefficient==0`` and ``exponent >= 0``, e.g. ``Decimal(0)``, then there is no precision
              beyond seconds.

            * After construction, ``microseconds``, ``fractional_precision``, and ``fractional_seconds``
              will all be present and normalized in the resulting :class:`Timestamp` instance.  If the
              precision of ``fractional_seconds`` is more than is capable of being expressed in
              ``microseconds``, then the ``microseconds`` field is truncated to six digits and
              ``fractional_precision`` is ``6``.

            Consider some examples:

            * `2019-10-01T12:45:01Z` would have the following fields set:
              * ``microseconds == 0``, ``fractional_precision == 0``, ``fractional_seconds == Decimal('0')``
            * `2019-10-01T12:45:01.100Z` would have the following fields set:
              * ``microseconds == 100000``, ``fractional_precision == 3``, ``fractional_seconds == Decimal('0.100')``
            * `2019-10-01T12:45:01.123456789Z` would have the following fields set:
              * ``microseconds == 123456``, ``fractional_precision == 6``, ``fractional_seconds ==
              Decimal('0.123456789')``

        Raises:
            ValueError: If any of the preconditions above are violated.
        """
    __slots__ = [TIMESTAMP_PRECISION_FIELD, TIMESTAMP_FRACTION_PRECISION_FIELD, TIMESTAMP_FRACTIONAL_SECONDS_FIELD]

    def __new__(cls, *args, **kwargs):
        def replace_microsecond(new_value):
            if has_microsecond_argument:
                lst = list(args)
                lst[DATETIME_CONSTRUCTOR_MICROSECOND_ARGUMENT_INDEX] = new_value
                return tuple(lst)
            else:
                kwargs[TIMESTAMP_MICROSECOND_FIELD] = new_value
                return args

        precision = None
        fractional_precision = None
        fractional_seconds = None
        datetime_microseconds = None
        has_microsecond_argument = len(args) > DATETIME_CONSTRUCTOR_MICROSECOND_ARGUMENT_INDEX
        if has_microsecond_argument:
            datetime_microseconds = args[DATETIME_CONSTRUCTOR_MICROSECOND_ARGUMENT_INDEX]
        elif TIMESTAMP_MICROSECOND_FIELD in kwargs:
            datetime_microseconds = kwargs.get(TIMESTAMP_MICROSECOND_FIELD)
        if TIMESTAMP_PRECISION_FIELD in kwargs:
            precision = kwargs.get(TIMESTAMP_PRECISION_FIELD)
            # Make sure we mask this before we construct the datetime.
            del kwargs[TIMESTAMP_PRECISION_FIELD]
        if TIMESTAMP_FRACTION_PRECISION_FIELD in kwargs:
            fractional_precision = kwargs.get(TIMESTAMP_FRACTION_PRECISION_FIELD)
            if fractional_precision is not None and not (0 <= fractional_precision <= MICROSECOND_PRECISION):
                raise ValueError('Cannot construct a Timestamp with fractional precision of %d digits, '
                                 'which is out of the supported range of [0, %d].'
                                 % (fractional_precision, MICROSECOND_PRECISION))
            # Make sure we mask this before we construct the datetime.
            del kwargs[TIMESTAMP_FRACTION_PRECISION_FIELD]
        if TIMESTAMP_FRACTIONAL_SECONDS_FIELD in kwargs:
            fractional_seconds = kwargs.get(TIMESTAMP_FRACTIONAL_SECONDS_FIELD)
            if fractional_seconds is not None:
                if not (0 <= fractional_seconds < 1):
                    raise ValueError('Cannot construct a Timestamp with fractional seconds of %s, '
                                     'which is out of the supported range of [0, 1).'
                                     % str(fractional_seconds))
            # Make sure we mask this before we construct the datetime.
            del kwargs[TIMESTAMP_FRACTIONAL_SECONDS_FIELD]

        if fractional_seconds is not None and (fractional_precision is not None or datetime_microseconds is not None):
            raise ValueError('fractional_seconds cannot be specified '
                             'when fractional_precision or microseconds are not None.')

        if fractional_precision is not None and datetime_microseconds is None:
            raise ValueError('datetime_microseconds cannot be None while fractional_precision is not None.')

        if fractional_precision == 0 and datetime_microseconds != 0:
            raise ValueError('datetime_microseconds cannot be non-zero while fractional_precision is 0.')

        if fractional_seconds is not None:
            fractional_seconds_exponent = fractional_seconds.as_tuple().exponent
            if fractional_seconds == DECIMAL_ZERO and fractional_seconds_exponent > 0:
                # Zero with a positive exponent is just zero. Set the exponent to zero so fractional_precision is
                # calculated correctly.
                fractional_seconds_exponent = 0
                fractional_seconds = DECIMAL_ZERO
            fractional_precision = min(-fractional_seconds_exponent, MICROSECOND_PRECISION)
            # Scale to microseconds and truncate to an integer.
            args = replace_microsecond(int(fractional_seconds * BASE_TEN_MICROSECOND_PRECISION_EXPONENTIATION))
        elif datetime_microseconds is not None:
            if fractional_precision is None:
                fractional_precision = MICROSECOND_PRECISION
            if fractional_precision == 0:
                # As previously verified, datetime_microseconds must be zero in this case.
                fractional_seconds = DECIMAL_ZERO
            else:
                try:
                    fractional_seconds = Decimal(datetime_microseconds).scaleb(-MICROSECOND_PRECISION)\
                        .quantize(PRECISION_LIMIT_LOOKUP[fractional_precision], context=Context(traps=[Inexact]))
                except Inexact:
                    raise ValueError('microsecond value %d cannot be expressed exactly in %d digits.'
                                     % (datetime_microseconds, fractional_precision))
        else:
            assert datetime_microseconds is None
            # The datetime constructor requires a non-None microsecond argument.
            args = replace_microsecond(0)
            fractional_precision = 0
            fractional_seconds = DECIMAL_ZERO

        instance = super(Timestamp, cls).__new__(cls, *args, **kwargs)
        setattr(instance, TIMESTAMP_PRECISION_FIELD, precision)
        setattr(instance, TIMESTAMP_FRACTION_PRECISION_FIELD, fractional_precision)
        setattr(instance, TIMESTAMP_FRACTIONAL_SECONDS_FIELD, fractional_seconds)

        return instance

    def __repr__(self):
        return 'Timestamp(%04d-%02d-%02dT%02d:%02d:%02d.%06d, %r, %r, %s=%s)' % \
               (self.year, self.month, self.day,
                self.hour, self.minute, self.second, self.microsecond,
                self.tzinfo, self.precision,
                TIMESTAMP_FRACTION_PRECISION_FIELD, self.fractional_precision)

    @staticmethod
    def adjust_from_utc_fields(*args, **kwargs):
        """Constructs a timestamp from UTC fields adjusted to the local offset if given."""
        raw_ts = Timestamp(*args, **kwargs)
        offset = raw_ts.utcoffset()
        if offset is None or offset == timedelta():
            return raw_ts

        # XXX This returns a datetime, not a Timestamp (which has our precision if defined)
        adjusted = raw_ts + offset
        if raw_ts.precision is None:
            # No precision means we can just return a regular datetime
            return adjusted

        return Timestamp(
            adjusted.year,
            adjusted.month,
            adjusted.day,
            adjusted.hour,
            adjusted.minute,
            adjusted.second,
            None,
            raw_ts.tzinfo,
            precision=raw_ts.precision,
            fractional_precision=None,
            fractional_seconds=raw_ts.fractional_seconds
        )


def timestamp(year, month=1, day=1,
              hour=0, minute=0, second=0, microsecond=None,
              off_hours=None, off_minutes=None,
              precision=None, fractional_precision=None, fractional_seconds=None):
    """Shorthand for the :class:`Timestamp` constructor.

    Specifically, converts ``off_hours`` and ``off_minutes`` parameters to a suitable
    :class:`OffsetTZInfo` instance.
    """
    delta = None
    if off_hours is not None:
        if off_hours < -23 or off_hours > 23:
            raise ValueError('Hour offset %d is out of required range -23..23.' % (off_hours,))
        delta = timedelta(hours=off_hours)
    if off_minutes is not None:
        if off_minutes < -59 or off_minutes > 59:
            raise ValueError('Minute offset %d is out of required range -59..59.' % (off_minutes,))
        minutes_delta = timedelta(minutes=off_minutes)
        if delta is None:
            delta = minutes_delta
        else:
            delta += minutes_delta

    tz = None
    if delta is not None:
        tz = OffsetTZInfo(delta)

    return Timestamp(
        year, month, day,
        hour, minute, second, microsecond,
        tz, precision=precision, fractional_precision=fractional_precision, fractional_seconds=fractional_seconds
    )


class Multimap(MutableMapping):
    """
    Dictionary that can hold multiple values for the same key

    In order not to break existing customers, getting and inserting elements with ``[]`` keeps the same behaviour
    as the built-in dict. If multiple elements are already mapped to the key, ``[]`  will return
    the newest one.

    To map multiple elements to a key, use the ``add_item`` operation.
    To retrieve all the values map to a key, use ``get_all_values``.
    """

    def __init__(self, *args, **kwargs):
        super(Multimap, self).__init__()
        self.__store = OrderedDict()
        if args is not None and len(args) > 0:
            for key, value in iter(args[0].items()):
                self.__store[key] = MultimapValue(value)

    def __getitem__(self, key):
        return self.__store[key][len(self.__store[key]) - 1]  # Return only one in order not to break clients

    def __delitem__(self, key):
        del self.__store[key]

    def __setitem__(self, key, value):
        self.__store[key] = MultimapValue(value)

    def __len__(self):
        return sum([len(values) for values in iter(self.__store.values())])

    def __iter__(self):
        for key in iter(self.__store.keys()):
            yield key

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return '{%s}' % ', '.join(['%r: %r' % (k, v) for k, v in self.items()])

    def add_item(self, key, value):
        if key in self.__store:
            self.__store[key].append(value)
        else:
            self.__setitem__(key, value)

    def get_all_values(self, key):
        return self.__store[key]

    def iteritems(self):
        for key in self.__store:
            for value in self.__store[key]:
                yield (key, value)

    def items(self):
        output = []
        for k, v in self.iteritems():
            output.append((k, v))
        return output


class MultimapValue(MutableSequence):

    def __init__(self, *args):
        if args is not None:
            self.__store = [x for x in args]
        else:
            self.__store = []

    def insert(self, index, value):
        self.__setitem__(index, value)

    def __len__(self):
        return len(self.__store)

    def __getitem__(self, index):
        return self.__store[index]

    def __setitem__(self, index, value):
        self.__store.insert(index, value)

    def __delitem__(self, index):
        del self.__store[index]

    def __iter__(self):
        for x in self.__store:
            yield x



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import amazon.ion.simpleion as ion


class IonLoadDump:
    """
    Wrapper for simpleion API that holds some configuration so that the configuration can be encapsulated here instead
    of having to be plumbed through the whole benchmark code.

    Results of profiling indicate that this adds a trivial amount of overhead, even for small data. If Ion Python
    performance improves by >1000% from June 2023, then this may need to be re-evaluated.
    """
    def __init__(self, binary, c_ext=True):
        self._binary = binary
        self._single_value = False
        # Need an explicit check here because if `None` is passed in as an argument, that is different from no argument,
        # and results in an unexpected behavior.
        self._c_ext = c_ext if c_ext is not None else True

    def loads(self, s):
        ion.c_ext = self._c_ext
        return ion.loads(s, single_value=self._single_value)

    def load(self, fp):
        ion.c_ext = self._c_ext
        return ion.load(fp, single_value=self._single_value)

    def dumps(self, obj):
        ion.c_ext = self._c_ext
        return ion.dumps(obj, binary=self._binary)

    def dump(self, obj, fp):
        ion.c_ext = self._c_ext
        return ion.dump(obj, fp, binary=self._binary)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

import pytest

from tests.event_aliases import *

from amazon.ion.util import coroutine
from amazon.ion.symbols import SymbolToken
from amazon.ion.simple_types import IonPySymbol
from amazon.ion.simpleion import _dump, loads, _FROM_TYPE


def _to_event_parameters():
    return [
        [loads('5'), [IonEvent(IonEventType.SCALAR, IonType.INT, 5, None, (), depth=0)]],
        [loads('abc'), [IonEvent(IonEventType.SCALAR, IonType.SYMBOL, SymbolToken('abc', None), None, (), depth=0)]],
        [loads('{abc: 1}'), [
            IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT, depth=0),
            IonEvent(IonEventType.SCALAR, IonType.INT, 1, SymbolToken('abc', None), (), depth=1),
            IonEvent(IonEventType.CONTAINER_END),
        ]],
        [loads('$0'), [IonEvent(IonEventType.SCALAR, IonType.SYMBOL, SymbolToken(None, 0), None, (), depth=0)]],
        [loads('{$0: $0}'), [
            IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT, depth=0),
            IonEvent(IonEventType.SCALAR, IonType.SYMBOL, IonPySymbol(None, 0), SymbolToken(None, 0), (), depth=1),
            IonEvent(IonEventType.CONTAINER_END),
        ]],
        [loads('[1, 2, 3, [4, 5, 6], [7, 8, 9]]'), [
            IonEvent(IonEventType.CONTAINER_START, IonType.LIST, depth=0),
              IonEvent(IonEventType.SCALAR, IonType.INT, 1, depth=1),
              IonEvent(IonEventType.SCALAR, IonType.INT, 2, depth=1),
              IonEvent(IonEventType.SCALAR, IonType.INT, 3, depth=1),
              IonEvent(IonEventType.CONTAINER_START, IonType.LIST, depth=1),
                IonEvent(IonEventType.SCALAR, IonType.INT, 4, depth=2),
                IonEvent(IonEventType.SCALAR, IonType.INT, 5, depth=2),
                IonEvent(IonEventType.SCALAR, IonType.INT, 6, depth=2),
              IonEvent(IonEventType.CONTAINER_END),
              IonEvent(IonEventType.CONTAINER_START, IonType.LIST, depth=1),
                IonEvent(IonEventType.SCALAR, IonType.INT, 7, depth=2),
                IonEvent(IonEventType.SCALAR, IonType.INT, 8, depth=2),
                IonEvent(IonEventType.SCALAR, IonType.INT, 9, depth=2),
              IonEvent(IonEventType.CONTAINER_END),
            IonEvent(IonEventType.CONTAINER_END),
        ]],

        [5, [IonEvent(IonEventType.SCALAR, IonType.INT, 5, None, (), depth=0)]],
        [u'abc', [IonEvent(IonEventType.SCALAR, IonType.STRING, "abc", None, (), depth=0)]],
        [{'abc': 1}, [
            IonEvent(IonEventType.CONTAINER_START, IonType.STRUCT, depth=0),
            IonEvent(IonEventType.SCALAR, IonType.INT, 1, "abc", (), depth=1),
            IonEvent(IonEventType.CONTAINER_END),
        ]],
    ]


def _to_event_test_name(params):
    return str(params[0])


@pytest.mark.parametrize("params", _to_event_parameters(), ids=_to_event_test_name)
def test_to_event(params):
    value, expected_events = params
    events = []

    @coroutine
    def event_receiver():
        event = yield
        while True:
            events.append(event)
            event = yield

    _dump(value, event_receiver(), _FROM_TYPE)

    assert events == expected_events




# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

from decimal import Decimal
from itertools import chain
from random import Random

from tests import parametrize, listify
from tests.reader_util import reader_scaffold, ReaderParameter, all_top_level_as_one_stream_params, value_iter
from tests.event_aliases import *

from amazon.ion.core import IonType, timestamp, TimestampPrecision, OffsetTZInfo
from amazon.ion.exceptions import IonException
from amazon.ion.reader import read_data_event, ReadEventType
from amazon.ion.reader_binary import _TypeID, _CONTAINER_TIDS, _TID_VALUE_TYPE_TABLE, stream_handler
from amazon.ion.symbols import SYMBOL_ZERO_TOKEN, SymbolToken

_PREC_YEAR = TimestampPrecision.YEAR
_PREC_MONTH = TimestampPrecision.MONTH
_PREC_DAY = TimestampPrecision.DAY
_PREC_MINUTE = TimestampPrecision.MINUTE
_PREC_SECOND = TimestampPrecision.SECOND

_ts = timestamp

_P = ReaderParameter

_IVM_PAIRS =[
    (NEXT, END),
    (e_read(b'\xE0\x01\x00\xEA'), IVM)
]


@listify
def _prepend_ivm(params):
    for p in params:
        new_p = _P(
            desc=p.desc,
            event_pairs=_IVM_PAIRS + p.event_pairs
        )
        yield new_p

_BASIC_PARAMS = (
    _P(
        desc='EMPTY',
        event_pairs=[(NEXT, END)]
    ),
    _P(
        desc='IVM',
        event_pairs=_IVM_PAIRS,
    ),
    _P(
        desc='IVM PARTS',
        event_pairs=[
            (NEXT, END),
            (e_read(b'\xE0\x01'), INC),
            (e_read(b'\x00\xEA'), IVM),
            (NEXT, END),
        ],
    ),
    _P(
        desc='IVM NOP IVM',
        event_pairs=[
            (NEXT, END),
            (e_read(b'\xE0\x01\x00\xEA\x02\x00'), IVM),
            (NEXT, INC),
            (e_read(b'\xFF\xE0\x01\x00\xEA'), IVM),
            (NEXT, END),
        ],
    ),
    _P(
        desc='SINGLE BYTE NOP',
        event_pairs=[
            (NEXT, END),
            (e_read(b'\xE0\x01\x00\xEA\x00'), IVM),
            (NEXT, END),
        ],
    ),
    _P(
        desc='LONG NOP',
        event_pairs=[
            (NEXT, END),
                                      # 0E is no-op with "l" 14 which means VarUint for length
            (e_read(b'\xE0\x01\x00\xEA\x0E' + \
                    b'\x8E' + (b'\x00' * 14)), IVM),
            (NEXT, END),
        ],
    ),
    _P(
        desc='NO START IVM',
        event_pairs=[
            (NEXT, END),
            (e_read(b'\x0F'), IonException),
        ],
    ),
)

_BAD_VALUES = (
    # The annotation wrapper declares 6 octets, but the wrapped value (a symbol value) ends after only 4.
    (b'\xe6\x81\x84\x71\x04\x71\x04', 'ANNOT LENGTH TOO LONG - SCALAR'),
    # The annotation wrapper declares 6 octets, but the wrapped value (a list) ends after only 4.
    (b'\xe6\x81\x84\xb1\x20\x71\x04', 'ANNOT LENGTH TOO LONG - CONTAINER'),
    # The annotation wrapper declares 4 octets, and the SIDs take up all four.
    (b'\xe4\x83\x84\x85\x86', 'ANNOT LENGTH TOO SHORT - NO VALUE'),
    # The annotation wrapper declares 4 octets, but the subfields (including a list) take up five.
    (b'\xe4\x81\x84\xb2\x21\x01', 'ANNOT LENGTH TOO SHORT - CONTAINER'),
    # The annotation wrapper declares 3 octets, but the subfields (including an int) take up four.
    (b'\xe3\x81\x84\x21\x01', 'ANNOT LENGTH TOO SHORT - SCALAR'),
    # TODO: annnotated nop is a fail

    # TODO: value within container is longer than container.
)


def _bad_params():
    for pair in _BAD_VALUES:
        yield _P(
            desc=pair[1],
            event_pairs=_IVM_PAIRS + [
                (NEXT, END),
                (e_read(pair[0]), IonException)
            ]
        )

# This is an encoding of a single top-level value and the expected events with ``NEXT``.
_TOP_LEVEL_VALUES = (
    (b'\x0F', e_null()),

    (b'\x10', e_bool(False)),
    (b'\x11', e_bool(True)),
    (b'\x1F', e_bool()),

    (b'\x2F', e_int()),
    (b'\x20', e_int(0)),
    (b'\x21\xFE', e_int(0xFE)),
    (b'\x22\x00\x01', e_int(1)),
    (b'\x24\x01\x2F\xEF\xCC', e_int(0x12FEFCC)),
    (b'\x29\x12\x34\x56\x78\x90\x12\x34\x56\x78', e_int(0x123456789012345678)),
    (b'\x2E\x81\x05', e_int(5)), # Over padded length.

    (b'\x3F', e_int()),  # null.int has two equivalent representations.
    (b'\x31\x01', e_int(-1)),
    (b'\x32\xC1\xC2', e_int(-0xC1C2)),
    (b'\x36\xC1\xC2\x00\x00\x10\xFF', e_int(-0xC1C2000010FF)),
    (b'\x39\x12\x34\x56\x78\x90\x12\x34\x56\x78', e_int(-0x123456789012345678)),
    (b'\x3E\x82\x00\xA0', e_int(-160)), # Over padded length + overpadded integer.

    (b'\x4F', e_float()),
    (b'\x40', e_float(0.0)),
    (b'\x44\x3F\x80\x00\x00', e_float(1.0)),
    (b'\x44\x7F\x80\x00\x00', e_float(float('+Inf'))),
    (b'\x48\x42\x02\xA0\x5F\x20\x00\x00\x00', e_float(1e10)),
    (b'\x48\x7F\xF8\x00\x00\x00\x00\x00\x00', e_float(float('NaN'))),

    (b'\x5F', e_decimal()),
    (b'\x50', e_decimal(Decimal())),
    (b'\x52\x47\xE8', e_decimal(Decimal('0e-1000'))),
    (b'\x54\x07\xE8\x00\x00', e_decimal(Decimal('0e1000'))),
    (b'\x52\x81\x01', e_decimal(Decimal('1e1'))),
    (b'\x53\xD4\x04\xD2', e_decimal(Decimal('1234e-20'))),
    (b'\x52\x80\x80', e_decimal(Decimal(0).copy_negate())),
    (b'\x52\x80\x01', e_decimal(Decimal('1e0'))),
    (b'\x52\xC1\x01', e_decimal(Decimal('1e-1'))),
    (b'\x51\xC1', e_decimal(Decimal('0e-1'))),
    (b'\x51\x81', e_decimal(Decimal('0e1'))),
    (b'\x52\x81\x81', e_decimal(Decimal('-1e1'))),
    (b'\x52\x80\x81', e_decimal(Decimal('-1e0'))),
    (b'\x52\xC1\x81', e_decimal(Decimal('-1e-1'))),
    (b'\x52\xC1\x80', e_decimal(Decimal('-0e-1'))),
    (b'\x52\x81\x80', e_decimal(Decimal('-0e1'))),

    (b'\x6F', e_timestamp()),
    (b'\x63\xC0\x0F\xE0', e_timestamp(_ts(2016, precision=_PREC_YEAR))), # -00:00
    (b'\x63\x80\x0F\xE0', e_timestamp(_ts(2016, off_hours=0, precision=_PREC_YEAR))),
    (
        b'\x64\x81\x0F\xE0\x82',
        e_timestamp(_ts(2016, 2, 1, 0, 1, off_minutes=1, precision=_PREC_MONTH))
    ),
    (
        b'\x65\xFC\x0F\xE0\x82\x82',
        e_timestamp(_ts(2016, 2, 1, 23, 0, off_hours=-1, precision=_PREC_DAY))
    ),
    (
        b'\x68\x43\xA4\x0F\xE0\x82\x82\x87\x80',
        e_timestamp(_ts(2016, 2, 2, 0, 0, off_hours=-7, precision=_PREC_MINUTE))
    ),
    (
        b'\x69\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E',
        e_timestamp(_ts(2016, 2, 2, 0, 0, 30, off_hours=-7, precision=_PREC_SECOND))
    ),
    (
        b'\x69\xC0\x81\x81\x81\x80\x80\x80\xC7\x01',
        e_timestamp(_ts(1, 1, 1, 0, 0, 0, None, precision=_PREC_SECOND, fractional_precision=None,
                        fractional_seconds=Decimal('1e-7')))
    ),
    (
        b'\x6B\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC3\x01',
        e_timestamp(_ts(
            2016, 2, 2, 0, 0, 30, 1000, off_hours=-7, precision=_PREC_SECOND, fractional_precision=3
        ))
    ),
    (
        b'\x67\xC0\x81\x81\x81\x80\x80\x80',
        e_timestamp(_ts(year=1, month=1, day=1, precision=_PREC_SECOND))
    ),
    (
        b'\x67\xC1\x81\x81\x81\x80\x81\x80',
        e_timestamp(_ts(year=1, month=1, day=1, off_minutes=-1, precision=_PREC_SECOND))
    ),
    (
        b'\x69\xC1\x81\x81\x81\x80\x81\x80\x80\x00',  # Fractions with coefficients of 0 and exponents > -1 are ignored.
        e_timestamp(_ts(year=1, month=1, day=1, off_minutes=-1, precision=_PREC_SECOND))
    ),
    (
        b'\x69\xC0\x81\x81\x81\x80\x80\x80\xC6\x01',
        e_timestamp(_ts(
            year=1, month=1, day=1, hour=0, minute=0, second=0, microsecond=1,
            precision=_PREC_SECOND, fractional_precision=6
        ))
    ),
    (
        b'\x6C\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC6\x03\xE8',  # The last three octets represent 1000d-6
        e_timestamp(_ts(
            2016, 2, 2, 0, 0, 30, 1000, off_hours=-7, precision=TimestampPrecision.SECOND
        )),
    ),
    (
        b'\x6C\x43\xA4\x0F\xE0\x82\x82\x87\x80\x9E\xC6\x03\xE8',  # The last three octets represent 1000d-6
        e_timestamp(_ts(
            2016, 2, 2, 0, 0, 30, 1000, off_hours=-7, precision=TimestampPrecision.SECOND, fractional_precision=6
        )),
    ),

    (b'\x7F', e_symbol()),
    (b'\x70', e_symbol(SYMBOL_ZERO_TOKEN)),
    (b'\x71\x02', e_symbol(SymbolToken(None, 2))),
    (b'\x7A' + b'\xFF' * 10, e_symbol(SymbolToken(None, 0xFFFFFFFFFFFFFFFFFFFF))),

    (b'\x8F', e_string()),
    (b'\x80', e_string(u'')),
    (b'\x84\xf0\x9f\x92\xa9', e_string(u'\U0001F4A9')),
    (b'\x88$ion_1_0', e_string(u'$ion_1_0')),

    (b'\x9F', e_clob()),
    (b'\x90', e_clob(b'')),
    (b'\x94\xf0\x9f\x92\xa9', e_clob(b'\xf0\x9f\x92\xa9')),

    (b'\xAF', e_blob()),
    (b'\xA0', e_blob(b'')),
    (b'\xA4\xf0\x9f\x92\xa9', e_blob(b'\xf0\x9f\x92\xa9')),

    (b'\xBF', e_null_list()),
    (b'\xB0', e_start_list(), e_end_list()),

    (b'\xCF', e_null_sexp()),
    (b'\xC0', e_start_sexp(), e_end_sexp()),

    (b'\xDF', e_null_struct()),
    (b'\xD0', e_start_struct(), e_end_struct()),
    (b'\xD1\x82\x84\x20', e_start_struct(), e_int(0, field_name=SymbolToken(None, 4)), e_end_struct())
)


def _top_level_event_pairs(data, events):
    first = True
    for event in events:
        input_event = NEXT
        if first:
            input_event = e_read(data)
            first = False
        yield input_event, event

_top_level_iter = partial(value_iter, _top_level_event_pairs, _TOP_LEVEL_VALUES)


def _gen_type_len(tid, length):
    """Very primitive type length encoder."""
    type_code = tid << 4
    if length < 0xE:
        return bytes((type_code | length,))
    else:
        type_code |= 0xE
        if length <= 0x7F:
            return bytes((type_code,)) + bytes((0x80 | length,))

    raise ValueError('No support for long lengths in reader test')

_TEST_ANNOTATION_DATA = b'\x82\x84\x87'
_TEST_ANNOTATION_LEN = len(_TEST_ANNOTATION_DATA)
_TEST_ANNOTATION_SIDS = (SymbolToken(None, 4), SymbolToken(None, 7))


def _top_level_value_params():
    """Converts the top-level tuple list into parameters with appropriate ``NEXT`` inputs.

    The expectation is starting from an end of stream top-level context.
    """
    for data, event_pairs in _top_level_iter():
        _, first = event_pairs[0]
        yield _P(
            desc='TL %s - %s - %r' % \
                 (first.event_type.name, first.ion_type.name, first.value),
            event_pairs=[(NEXT, END)] + event_pairs + [(NEXT, END)],
        )


def _annotate_params(params):
    """Adds annotation wrappers for a given iterator of parameters,

    The requirement is that the given parameters completely encapsulate a single value.
    """
    for param in params:
        @listify
        def annotated():
            for input_event, output_event in param.event_pairs:
                if input_event.type is ReadEventType.DATA:
                    data_len = _TEST_ANNOTATION_LEN + len(input_event.data)
                    data = _gen_type_len(_TypeID.ANNOTATION, data_len) \
                           + _TEST_ANNOTATION_DATA \
                           + input_event.data
                    input_event = read_data_event(data)
                    output_event = output_event.derive_annotations(_TEST_ANNOTATION_SIDS)
                yield input_event, output_event

        yield _P(
            desc='ANN %s' % param.desc,
            event_pairs=annotated(),
        )


def _data_event_len(event_pairs):
    length = 0
    for read_event, _ in event_pairs:
        if read_event.type is ReadEventType.DATA:
            length += len(read_event.data)
    return length


def _containerize_params(params, with_skip=True):
    """Adds container wrappers for a given iteration of parameters.

    The requirement is that each parameter is a self-contained single value.
    """
    rnd = Random()
    rnd.seed(0xC0FFEE)
    params = list(params)
    for param in params:
        data_len = _data_event_len(param.event_pairs)
        for tid in _CONTAINER_TIDS:
            ion_type = _TID_VALUE_TYPE_TABLE[tid]

            field_data = b''
            field_tok = None
            field_desc = ''
            if ion_type is IonType.STRUCT:
                field_sid = rnd.randint(0, 0x7F)
                field_data = bytes((field_sid | 0x80,))
                field_tok = SymbolToken(None, field_sid)
                field_desc = ' (f:0x%02X)' % field_sid

            @listify
            def add_field_names(event_pairs):
                first = True
                for read_event, ion_event in event_pairs:
                    if first and not ion_event.event_type.is_stream_signal:
                        ion_event = ion_event.derive_field_name(field_tok)
                        first = False
                    yield read_event, ion_event

            type_header = _gen_type_len(tid, data_len + len(field_data)) + field_data

            start = [
                (NEXT, END),
                (e_read(type_header), e_start(ion_type)),
                (NEXT, INC)
            ]
            mid = add_field_names(param.event_pairs[1:-1])
            end = [(NEXT, e_end(ion_type)), (NEXT, END)]

            desc = 'SINGLETON %s%s - %s' % (ion_type.name, field_desc, param.desc)
            yield _P(
                desc=desc,
                event_pairs=start + mid + end,
            )

            # Version with SKIP
            if with_skip:
                @listify
                def only_data_inc(event_pairs):
                    for read_event, ion_event in event_pairs:
                        if read_event.type is ReadEventType.DATA:
                            yield read_event, INC

                start = start[:-1] + [(SKIP, INC)]
                end = only_data_inc(param.event_pairs)
                end = end[:-1] + [(end[-1][0], e_end(ion_type)), (NEXT, END)]
                
                yield _P(
                    desc='SKIP %s' % desc,
                    event_pairs=start + end,
                )

# TODO Add NOP pad fuzz.
# TODO Add data incomplete fuzz.


@parametrize(*chain(
    _BASIC_PARAMS,
    _bad_params(),
    _prepend_ivm(_top_level_value_params()),
    _prepend_ivm(_annotate_params(_top_level_value_params())),
    _prepend_ivm(_containerize_params(_top_level_value_params())),
    _prepend_ivm(
        _containerize_params(
            _containerize_params(_top_level_value_params(), with_skip=False)
        )
    ),
    _prepend_ivm(all_top_level_as_one_stream_params(_top_level_iter)),
))
def test_raw_reader(p):
    reader_scaffold(stream_handler(), p.event_pairs)



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at:
#
#    http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the
# License.

"""Buffer for binary Ion writers."""


class _Node:
    def __init__(self, value=None):
        self.value = value
        self.parent = None
        self.children = None

    def add_child(self, node):
        if self.children is None:
            self.children = []
        node.parent = self
        self.children.append(node)

    def add_leaf(self, node):
        if not self.children:
            self.add_child(node)
            return
        leftmost_child = self.children[0]
        while leftmost_child.children:
            leftmost_child = leftmost_child.children[0]
        leftmost_child.add_child(node)


class BufferTree:
    """A tree of buffers that can be depth-first traversed to produce a
    correctly-ordered Ion stream.

    Containers are treated as subtrees. Calling ``start_container`` creates
    an empty node which represents the start of a new subtree, while calling
    ``end_container`` steps out of that subtree, allowing for additional
    nodes to be added as siblings of that subtree's start node.
    It is important that there is exactly one call to start_container per call
    to ``end_container``. Nodes representing the start of a container subtree
    are kept in a stack; upon calling ``end_container``, the subtree represented
    by the node that was created by the most recent call to start_container will
    be stepped out. Other than that, no container semantics are asserted by this class.

    Scalar values are treated as child nodes of the current container subtree
    (which may be at the top-level if the current container subtree's root node
    is the root node of the BufferTree). Calling add_scalar_value will add a new
    node as a child of the current container subtree. A container subtree's
    children are ordered.

    If None values are passed in to end_container or add_scalar_value,
    nodes will be added to the tree, in the same way as described above, but no value
    will be yielded for that node upon drain.
    """
    def __init__(self):
        self.__root = None
        self.__container_lengths = None  # Stack of pending container lengths.
        self.__container_nodes = None  # Stack of pending container nodes.
        self.__container_node = None  # Node representing the currently active container.
        self.current_container_length = None  # Length of the currently active container.
        self.__reset()

    def __reset(self):
        self.__root = _Node()
        self.__container_lengths = []
        self.__container_nodes = []
        self.__container_node = self.__root
        self.current_container_length = 0

    def __depth_traverse(self, node):
        if node.children:
            for child in node.children:
                for val in self.__depth_traverse(child):
                    yield val
        yield node.value

    def start_container(self):
        """Add a node to the tree that represents the start of a container.

        Until end_container is called, any nodes added through add_scalar_value
        or start_container will be children of this new node.
        """
        self.__container_lengths.append(self.current_container_length)
        self.current_container_length = 0
        new_container_node = _Node()
        self.__container_node.add_child(new_container_node)
        self.__container_nodes.append(self.__container_node)
        self.__container_node = new_container_node

    def end_container(self, header_buf):
        """Add a node containing the container's header to the current subtree.

        This node will be added as the leftmost leaf of the subtree that was
        started by the matching call to start_container.

        Args:
            header_buf (bytearray): bytearray containing the container header.
        """
        if not self.__container_nodes:
            raise ValueError("Attempted to end container with none active.")
        # Header needs to be the first node visited on this subtree.
        self.__container_node.add_leaf(_Node(header_buf))
        self.__container_node = self.__container_nodes.pop()
        parent_container_length = self.__container_lengths.pop()
        self.current_container_length = \
            parent_container_length + self.current_container_length + len(header_buf)

    def add_scalar_value(self, value_buf):
        """Add a node to the tree containing a scalar value.

        Args:
            value_buf (bytearray): bytearray containing the scalar value.
        """
        self.__container_node.add_child(_Node(value_buf))
        self.current_container_length += len(value_buf)

    def drain(self):
        """Walk the BufferTree and reset it when finished.

        Yields:
            any: The current node's value.
        """
        if self.__container_nodes:
            raise ValueError("Attempted to drain without ending all containers.")
        for buf in self.__depth_traverse(self.__root):
            if buf is not None:
                yield buf
        self.__reset()

